package ch.thomson.ucs.alarms;

import java.util.LinkedList;
import java.util.List;

import ch.thomson.ucs.alarms.AlarmVectorHandler.Category;
import ch.thomson.ucs.alarms.AlarmVectorHandler.Descriptor;
import ch.thomson.ucs.alarms.client.ClientUtil;
import ch.thomson.ucs.alarms.client.dao.AlarmHistoryDAO;
import ch.thomson.ucs.alarms.client.dao.AlarmMessageDAO;
import ch.thomson.ucs.alarms.client.dao.AlarmPendingDAO;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmHistoryEntry;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;

public class AlarmVectorHandlerFactory implements ResetListener, AlarmSumHandler {

	private static AlarmVectorHandlerFactory instance;
	private List<AlarmVectorHandler> handlers = new LinkedList<AlarmVectorHandler>();
	private List<WarningAlarmVectorHandler> warningHandlers = new LinkedList<WarningAlarmVectorHandler>();
	private StatemachineAlarmVectorHandler statemachineHandler;
	private PropertyNumericDAO sumWarning;
	private PropertyNumericDAO sumTrip;
	private PropertyNumericDAO sumFault;
	private PropertyNumericDAO alarmVector;
	private AlarmPendingDAO alarmPendingDAO;
	private AlarmHistoryDAO alarmHistoryDAO;
	public static final String MESSAGE_RESET_FAULTS = "M1023";
	
	
	private AlarmVectorHandlerFactory() throws Exception {
		setupMessages();
		ResetHandler.instance().addResetListener(this);
	}
	
	public void setStatemachineVectorHandler(StatemachineAlarmVectorHandler handler) {
		if(this.statemachineHandler != null) {
			handlers.remove(statemachineHandler);
		}
		
		this.statemachineHandler = handler;
		handlers.add(handler);
	}
	
	private void setSumWarning(boolean state) {
		sumWarning.setInt(state ? 1 : 0);
	}
	
	private void setSumTrip(boolean state) {
		sumTrip.setInt(state ? 1 : 0);
	}
	
	private void setSumFault(boolean state) {
		sumFault.setInt(state ? 1 : 0);
	}
	
	public AlarmPendingDAO getAlarmPendingDAO() {
		if(alarmPendingDAO == null)
			alarmPendingDAO = ClientUtil.createAlarmPendingDAO();
		
		if(alarmPendingDAO.isClosed())
			try {
				alarmPendingDAO.open("localhost");
			} catch (Exception e) {
				e.printStackTrace();
			}
		
		return alarmPendingDAO;
	}
	
	public AlarmHistoryDAO getAlarmHistoryDAO() {
		if(alarmHistoryDAO == null)
			alarmHistoryDAO = ClientUtil.createAlarmHistoryDAO();
		
		if(alarmHistoryDAO.isClosed())
			try {
				alarmHistoryDAO.open("localhost");
			} catch (Exception e) {
				e.printStackTrace();
			}
		
		return alarmHistoryDAO;
	}
	
	public void setSumWarningProperty(String path) throws Exception {
		sumWarning = PropertyNumericDAO.createPropertyNumericDAO(path, true);
	}
	
	public void setSumTripProperty(String path) throws Exception {
		sumTrip = PropertyNumericDAO.createPropertyNumericDAO(path, true);
	}
	
	public void setSumFaultProperty(String path) throws Exception {
		sumFault = PropertyNumericDAO.createPropertyNumericDAO(path, true);
	}
	
	public void setAlarmVector(String path) throws Exception {
		alarmVector = PropertyNumericDAO.createPropertyNumericDAO(path, true);
	}
	
	protected PropertyNumericDAO getAlarmVector() {
		return alarmVector;
	}
	
	public AlarmVectorHandler createAlarmVectorHandler(String propertyPath, int messageOffset, Category category, Descriptor descriptor) throws Exception {
		PropertyNumericDAO p = PropertyNumericDAO.createPropertyNumericDAO(propertyPath, true);
		
		AlarmVectorHandler handler = null;
		
		if(Category.WARNING.equals(category)) {
			handler = new WarningAlarmVectorHandler(p, messageOffset, descriptor);
			warningHandlers.add((WarningAlarmVectorHandler) handler);
		}
		
		if(Category.FAULT.equals(category) && (Descriptor.TRIP.equals(descriptor) || Descriptor.FIRST_TRIP.equals(descriptor)))
			handler = new TripAlarmVectorHandler(p, messageOffset, descriptor);
		
		if(Category.FAULT.equals(category) && (Descriptor.FAULT.equals(descriptor) || Descriptor.FIRST_FAULT.equals(descriptor)))
			handler = new FaultAlarmVectorHandler(p, messageOffset, descriptor);
		
		if(Category.FAULT.equals(category) && Descriptor.FATAL_ERROR.equals(descriptor))
			handler = new FaultAlarmVectorHandler(p, messageOffset, descriptor);
		
		if(handler == null)
			throw new Exception("Unknown Category of Alarm Vector Handler");
		
		handlers.add(handler);
		return handler;
	}
	
	public static synchronized AlarmVectorHandlerFactory getInstance() throws Exception {
		if(instance == null)
			instance = new AlarmVectorHandlerFactory();
		return instance;
	}
	
	private void setupMessages() throws Exception {
		
		AlarmMessageDAO dao = ClientUtil.createAlarmMessageDAO();
		String[] idList = Alarms.getIds();
		try {
			dao.open("localhost");
			dao.setAutoCommit(false);
			for (String id : idList) {
				dao.setEntry(id, Alarms.getString(id));
			}
			dao.commit();
		} finally {
			dao.setAutoCommit(true);
			dao.close();
		}
	}

	@Override
	public void resetAlarms(boolean reset) {
		if(reset) {
			this.setSumWarning(false);
			this.setSumTrip(false);
			this.setSumFault(false);
			
			try {
				this.getAlarmPendingDAO().deleteAllEntries();
				this.getAlarmHistoryDAO().setEntry(new AlarmHistoryEntry(System.currentTimeMillis(), MESSAGE_RESET_FAULTS, "System", Category.MESSAGE.getName(), Descriptor.NO.getName()));
			} catch (Exception e) {
				this.getAlarmPendingDAO().close();
			}
			
			for(AlarmVectorHandler h : handlers) 
				h.reset();
		}
	}

	@Override
	public void setSumWarning() {
		this.setSumWarning(true);
	}

	@Override
	public void setSumTrip() {
		this.setSumTrip(true);
	}

	@Override
	public void setSumFault() {
		this.setSumFault(true);
	}

	@Override
	public void warningIsGone() {
		boolean pendingWarnings = false;
		
		// berprfe, ob es noch andere Warnungen hat
		for(WarningAlarmVectorHandler wh : warningHandlers) {
			pendingWarnings |= wh.hasPendingAlarms();
		}
		
		if(this.statemachineHandler != null)
			pendingWarnings |= this.statemachineHandler.hasPendingWarnings();
		
		this.setSumWarning(pendingWarnings);
	}
	
	public void setAlarmVectorValueMasked(int index, int mask) {
		synchronized(alarmVector) {
			int act = alarmVector.getInt(index);
			int oldAct = act;
			act |= mask;
			if(oldAct != act) {
				alarmVector.setInt(index, act);
			}
		}
	}
	
	public void resetAlarmVectorValueMasked(int index, int mask) {
		synchronized(alarmVector) {
			int act = alarmVector.getInt(index);
			int oldAct = act;
			act &= ~mask;
			if(oldAct != act) {
				alarmVector.setInt(index, act);
			}
		}
	}
	
	public void resetAlarmVector() {
		synchronized(alarmVector) {
			alarmVector.setValuesAsInt(new int[alarmVector.getLength()]);
		}
	}
	
	public void setAlarmVectorValue(int index, int value) {
		synchronized(alarmVector) {
			alarmVector.setInt(index, value);
		}
	}
}
