package ch.thomson.ucs.alarms.client.dao.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.LinkedList;
import java.util.List;

import ch.thomson.ucs.alarms.client.dao.AlarmHistoryDAO;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmHistoryEntry;

public class AlarmHistoryDAOImpl extends BasicDAOImpl implements AlarmHistoryDAO {
	private int maxEntries = 20000;
		
	public void setupTable() throws SQLException {
		StringBuilder query = new StringBuilder();
		query.append("CREATE TABLE IF NOT EXISTS alarms_history (");
		query.append("timestamp TIMESTAMP NOT NULL, ");
		query.append("id VARCHAR(16) NOT NULL, ");
		query.append("eqm VARCHAR(32) NOT NULL, ");
		query.append("category VARCHAR(16) NOT NULL, ");
		query.append("descriptor VARCHAR(16), ");
		query.append("CONSTRAINT ah_pk PRIMARY KEY (timestamp, id), ");
		query.append("FOREIGN KEY (id) REFERENCES messages);");
		
		Statement st = null;
		try {
			st = connection.createStatement();
			st.executeUpdate(query.toString());
		} finally {
			if(st != null)
				st.close();
		}
	}
	
	private int getNumberOfEntries() throws SQLException {
		int result = 0;
		String sql = "SELECT COUNT(*) FROM alarms_history;";
		Statement st = null;
		try {
			st = connection.createStatement();
			ResultSet rs =st.executeQuery(sql);
			if(rs.next())
				result = rs.getInt(1);
		} finally {
			if(st != null)
				st.close();
		}
		return result;
	}
	
	private void cleanUp() throws SQLException {
		int nEntriesToDelete = getNumberOfEntries() - maxEntries;
		if(nEntriesToDelete <= 0)
			return;
		
		String sql = "DELETE FROM alarms_history WHERE timestamp IN (SELECT timestamp FROM alarms_history ORDER BY timestamp ASC LIMIT ?);";
		PreparedStatement ps = null;
		try {
			ps = connection.prepareStatement(sql);
			ps.setInt(1, nEntriesToDelete);
			ps.executeUpdate();
		} finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public synchronized void setEntry(AlarmHistoryEntry entry) throws Exception {
		String insert = "INSERT INTO alarms_history VALUES (?, ?, ?, ?, ?);";
		PreparedStatement ps = null;
		try {
			connection.setAutoCommit(false);
			ps = connection.prepareStatement(insert);
			ps.setTimestamp(1, entry.getTimestamp());
			ps.setString(2, entry.getId());
			ps.setString(3, entry.getEqm());
			ps.setString(4, entry.getCategory());
			ps.setString(5, entry.getDescriptor());
			ps.executeUpdate();
			cleanUp();
			connection.commit();
			
		} catch (SQLIntegrityConstraintViolationException e) { 
			// Nichts tun, Eintrag ist schon vorhanden
		} finally {
			if(ps != null)
				ps.close();
			connection.setAutoCommit(true);
		}
	}

	@Override
	public void deleteEntry(String id, String category) throws Exception {
		String delete = "DELETE FROM alarms_history WHERE id=? AND category=?;";
		PreparedStatement ps = null;
		try {
			ps = connection.prepareStatement(delete);
			ps.setString(1, id);
			ps.setString(2, category);
			ps.executeUpdate();
		}
		finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public void deleteAllEntries() throws Exception {
		Statement st = null;
		try {
			st = connection.createStatement();
			st.execute("DELETE FROM alarms_history;");
		} finally {
			if(st != null)
				st.close();
		}
	}

	@Override
	public List<AlarmHistoryEntry> getAllEntries() throws Exception {
		
		List<AlarmHistoryEntry> result = new LinkedList<AlarmHistoryEntry>();
		
		Statement st = null;
		
		try {
			st = connection.createStatement();
			ResultSet rs = st.executeQuery("SELECT * FROM alarms_history AS a, messages AS m WHERE a.id = m.id ORDER BY timestamp DESC;");
		
			while (rs.next()) {
				Timestamp timestamp = rs.getTimestamp("timestamp"); 	
				String id = rs.getString("id");
				String eqm = rs.getString("eqm");
				String category = rs.getString("category");
				String descriptor = rs.getString("descriptor");
				String message = rs.getString("message");
			
				AlarmHistoryEntry ahe = new AlarmHistoryEntry(timestamp, id, eqm, category, descriptor, message);
				result.add(ahe); 	
			}
		} finally {
			if(st != null)
				st.close();
		}
		
		return result;
	}

	public int getMaxEntries() {
		return maxEntries;
	}

	public void setMaxEntries(int maxEntries) {
		this.maxEntries = maxEntries;
	}
}
