package ch.thomson.ucs.desy.specialfunctions.fastchannel;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import ch.thomson.ucs.desy.specialfunctions.SpecialFunctionDeviceUpdated;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericChangeListener;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericNotification;

/**
 * Diese Klasse verwaltet die verschiedenen FastChannelCalculation und sorgt
 * dafr, dass gemeinsame Resourcen (z. B. die Sample-Arrays) gecached werden.
 * Dies steigert die Performance, da die jeweiligen Resourcen pro
 * Berechnungszyklus nur einmal gelesen werden mssen. Die Klasse wird vom Device
 * synchron informiert, unmittelbar nachdem die Daten vom Device gelesen wurden.
 * 
 * @author FischerU
 * 
 */
public class FastChannelCalculationSet extends SpecialFunctionDeviceUpdated {
	
	/**
	 * Aus Performance-Grnden werden die Daten der PropertyNumericDAO gecached,
	 * damit nicht jede FastCalculation wieder ein Lesen auslst.
	 * 
	 * @author FischerU
	 * 
	 */
	public interface CachedDAO {
		public void initialize();
		public void dispose();
		public void refresh();
	}
	
	public class Sample implements CachedDAO{
		private PropertyNumericDAO dao;
		private float[] cachedValues;
		
		public Sample(PropertyNumericDAO dao) {
			this.dao = dao;
		}
		
		@Override
		public void refresh() {
			cachedValues = dao.getValuesAsFloat();
		}

		public float[] getCachedValues() {
			return cachedValues;
		}

		@Override
		public void initialize() {}

		@Override
		public void dispose() {}
	}
	
	/**
	 * Diese Klasse wird beispielsweise fr die Marker bei der Bereichsauswahl
	 * der Average-Berechnung benutzt. Wenn der Marker ndert, muss ein Refresh
	 * ausgelst werden.
	 * 
	 * @author FischerU
	 * 
	 */
	public class Selection implements CachedDAO, PropertyNumericChangeListener {
		private PropertyNumericDAO dao;
		private int cachedValueInt;
		private float cachedValueFloat;
		
		public Selection(PropertyNumericDAO dao) {
			this.dao = dao;
		}
		
		@Override
		public void refresh() {
			cachedValueInt = dao.getInt();
			cachedValueFloat = dao.getFloat();
		}

		public int getCachedValueInt() {
			return cachedValueInt;
		}
		
		public float getCachedValueFloat() {
			return cachedValueFloat;
		}
		
		@Override
		public void initialize() {
			dao.addChangeListener(this);
		}
		
		@Override
		public void dispose() {
			dao.removeChangeListener(this);
		}

		@Override
		public void notifyChanged(PropertyNumericNotification msg) {
			FastChannelCalculationSet.this.refresh();
		}
	}
	
	public class Result {
		private PropertyNumericDAO dao;
		
		public Result(PropertyNumericDAO dao) {
			this.dao = dao;
			dao.getDevice().findDataTable(dao.getProperty().getDataTableName()).setObserverNotificationEnabled(false);
		}
		
		public void setResult(float result) {
			this.dao.setFloat(result);
		}
		
		public void setResult(float[] result) {
			this.dao.setValuesAsFloat(result);
		}
		
		public int getLength() {
			return dao.getLength();
		}
	}
	
	private Set<FastChannelCalculation> calculations = new HashSet<FastChannelCalculation>();
	private Map<String, CachedDAO> cachedDAOs = new HashMap<String, CachedDAO>();
	
	@Override
	public void initialize() {
		refresh();
		
		for(Map.Entry<String, CachedDAO> entry : cachedDAOs.entrySet())
			entry.getValue().initialize();
		
		super.initialize();
	}

	@Override
	public void postUpdate() {
		refresh();
	}
	
	private void refresh() {
		// Alle DAOs updaten
		for(Map.Entry<String, CachedDAO> entry : cachedDAOs.entrySet()) {
			//long time = System.nanoTime();
			entry.getValue().refresh();
			//System.out.println(entry.getValue().getClass() + ": " + (((System.nanoTime() - time) / 1000) / 1000.0));
		}
				
		// Alle Berechnungen updaten
		for(FastChannelCalculation calculation : calculations) {
			//long time = System.nanoTime();
			calculation.refresh();
			//System.out.println(calculation.getClass() + ": " + (((System.nanoTime() - time) / 1000) / 1000.0));
		}
	}

	@Override
	public String getName() {
		return this.getClass().getName();
	}
	
	public boolean addFastChannelCalculation(FastChannelCalculation fcc) {
		return this.calculations.add(fcc);
	}
	
	public boolean removeFastChannelCalculation(FastChannelCalculation fcc) {
		return this.removeFastChannelCalculation(fcc);
	}
	
	public Sample findSample(String path) throws Exception {
		CachedDAO dao = cachedDAOs.get(path);
		if(dao == null) {
			dao = new Sample(this.findProperty(path, false));
			cachedDAOs.put(path, dao);
		}
		return (Sample) dao;
	}
	
	public Selection findSelection(String path) throws Exception {
		CachedDAO dao = cachedDAOs.get(path);
		if(dao == null) {
			dao = new Selection(this.findProperty(path, true));
			cachedDAOs.put(path, dao);
		}
		return (Selection) dao;
	}
	
	public Result createResult(String path) throws Exception {
		return new Result(this.findProperty(path, false));
	}

	@Override
	public void dispose() {
		
		super.dispose();
		
		for(Map.Entry<String, CachedDAO> entry : cachedDAOs.entrySet())
			entry.getValue().dispose();
		
		Iterator<FastChannelCalculation> it = calculations.iterator();
		while(it.hasNext())
			it.next().dispose();
	}	
}
