package ch.thomson.ucs.ftdi;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;



/**
 * Stores all data, describing a particular device. This data is stored within
 * the EEPROM and can be written as well. For a detailed description of the date
 * see FTD`s publications. 
 */
public class Descriptor {

	final static short OFFSET_USER_AREA = 0xDC;		// -> 220 -> 0xDC
	final static short OFFSET_STRING_AREA = 0x9A;	// -> 0x9A
	
	// General Eeprom Area
	private byte portA = 8;				// adr=0x00 len=1	(8->SPI, 9->async FIFO)
	private byte portB = 9;				// adr=0x01 len=1	(8->SPI, 9->async FIFO)
	private short vendorId = 0x403;		// adr=0x02 len=2
	private short productId = 0x6010;	// adr=0x04 len=2	
	private short chipType = 0x700;		// adr=0x06 len=2	
	private byte configDesc = -96;		// adr=0x08 len=1	(default 0xa0)
									// 		Bit 7: always 1
									// 		Bit 6: 1 if is self powered, 0 if bus powered
									// 		Bit 5: 1 if uses remote wakeup
									// 		Bit 4: 1 if is battery powered	
	private byte maxPower = 0x4b;		// adr=0x09 len=1	(default max power 150mA)	
	private short chipConfig = 0x8;		// adr=0x0A len=1
									// 		Bit 4: 1 change USB version
									// 		Bit 3: 1 use the serial number string
									// 		Bit 2: 1 enable suspend pull downs
									// 		Bit 1: 1 out endpoint is isochronous
									// 		Bit 0: 1 in endpoint is isochronous
	
									// adr=0x0B len=1 -> always 0
	
	private short usbVersion = 0x1111;		// adr=0x0C len=2	
	private byte offsetManufacturer = -102;	// adr=0x0E len=1 (hard coded for 2232H=0x9A)	
	private byte lenManufacturer = 16;		// adr=0x0F len=1	
	private byte offsetProduct = 0;		// adr=0x10 len=1	
	private byte lenProduct = 0;			// adr=0x11 len=1	
	private byte offsetSerial = 0;		// adr=0x12 len=1	
	private byte lenSerial = 0;			// adr=0x13 len=1	
	private String manufacturer = "Thomson";	// adr=0x9A len=16	
	private String product = "Board";			// adr=0xAA len=32	
	private String serial = "TBAG0000";			// adr=0xCA len=18
	// User Eeprom Area
	private short boardId = 0;			// adr=0xDC len=2
	private int boardVersion = 0;		// adr=0xE0 len=4
	private int boardSerial = 0;		// adr=0xE4 len=4
	// General Eeprom Area
	private int checksum = 0xAAAA;		// adr=0xFE len=2
	
	private byte[] eepromBuffer = null;	// byte array for Eeprom buffer
	private byte[] userDataBuffer = null;	// byte array for the User Data 24 Byte

	/**
	 * @return the userDataBuffer
	 */
	public byte[] getUserDataBuffer() {
		return userDataBuffer;
	}

	/**
	 * @param userDataBuffer the userDataBuffer to set
	 */
	public void setUserDataBuffer(byte[] userDataBuffer) {
		this.userDataBuffer = userDataBuffer;
	}

	/**
	 * Creates a new <code>Descriptor</code> for writing. 
	 * 
	 */
	public Descriptor() {
	}
	
	/**
	 * Creates a new <code>Descriptor</code> form read buffer. 
	 * 
	 */
	public Descriptor(byte[] buffer) throws IllegalArgumentException {
		if (buffer == null)
			throw new IllegalArgumentException("The buffer is empty.");

		if (buffer.length != 256)
			throw new IllegalArgumentException("The buffer is not 256 byte long.");
		
		eepromBuffer = buffer;
		parseBuffer();
	}
	
	/**
	 * return Eeprom Buffer. 
	 * 
	 */
	public byte [] getEepromBuffer() {
		buildBuffer();
		return eepromBuffer;
	}

	/**
	 * return Port mode SPI / FIFO. 
	 * 
	 */
	public byte getPortMode(int index) {
		if(index == 0)
			return portA;
		else
			return portB;
	}
	/**
	 * set Port mode SPI / FIFO. 
	 * 
	 */
	public void setPortMode(int index, byte val) {
		if(index == 0)
			portA = val;
		else
			portB = val;
	}

	/**
	 * return Product String (e.g. PSM) 
	 * 
	 */
	public String getProduct() {
		return product;
	}
	/**
	 * set Product String (e.g. PSM) 
	 * 
	 */
	public void setProduct(String val) {
		product = val;
	}

	/**
	 * return Self Powered. 
	 * 
	 */
	public boolean getSelfPowered() {
		boolean ret = false;
		// configDesc = -96;		// adr=0x08 len=1	(default 0xa0)
		// 		Bit 7: always 1
		// 		Bit 6: 1 if is self powered, 0 if bus powered
		// 		Bit 5: 1 if uses remote wakeup
		// 		Bit 4: 1 if is battery powered	
		if((configDesc & 0x40) == 0x40)
			ret = true;
		return ret;
	}
	/**
	 * set Self Powered. 
	 * 
	 */
	public void setSelfPowered(boolean val) {
		if(val == true)
			configDesc |= 0x40;
		else
			configDesc &= 0xBF;
	}

	/**
	 * return User Area Board ID. 
	 * 
	 */
	public short getUABoardId() {
		return boardId;
	}
	/**
	 * set User Area Board ID. 
	 * 
	 */
	public void setUABoardId(short val) {
		boardId = val;
	}
	
	/**
	 * return User Area Board Version. 
	 * 
	 */
	public int getUABoardVersion() {
		return boardVersion;
	}
	/**
	 * set User Area Board Version. 
	 * 
	 */
	public void setUABoardVersion(int val) {
		boardVersion = val;
	}
	
	/**
	 * return User Area Serial Number. 
	 * 
	 */
	public int getUABoardSerial() {
		return boardSerial;
	}
	/**
	 * set User Area Serial Number. 
	 * 
	 */
	public void setUABoardSerial(int val) {
		boardSerial = val;
	}
	
	/**
	 * parse the eeprom buffer. 
	 * 
	 */
	private void parseBuffer() throws IllegalArgumentException {
		
		if(eepromBuffer == null)
			throw new IllegalArgumentException("The buffer is empty.");
		
		// check first if buffer is empty no write is done before
		boolean empty = true;
		for(byte b : eepromBuffer) {
			if(b != -1)
				empty = false;
		}
		if(empty)
			return;

		ByteBuffer bb = ByteBuffer.wrap(eepromBuffer);
		bb.order(ByteOrder.LITTLE_ENDIAN);
	
		portA = bb.get();							// adr=0x00 len=1	(8->SPI, 9->async FIFO)
		portB = bb.get();							// adr=0x01 len=1	(8->SPI, 9->async FIFO)
		vendorId = bb.getShort();					// adr=0x02 len=2
		productId = bb.getShort();					// adr=0x04 len=2	
		chipType = bb.getShort();					// adr=0x06 len=2	
		configDesc = bb.get();						// adr=0x08 len=1
													// 		Bit 7: always 1
													// 		Bit 6: 1 if is self powered, 0 if bus powered
													// 		Bit 5: 1 if uses remote wakeup
													// 		Bit 4: 1 if is battery powered	
		maxPower = bb.get();						// adr=0x09 len=1	
		chipConfig = bb.getShort();					// adr=0x0A len=1
													// 		Bit 4: 1 change USB version
													// 		Bit 3: 1 use the serial number string
													// 		Bit 2: 1 enable suspend pull downs
													// 		Bit 1: 1 out endpoint is isochronous
													// 		Bit 0: 1 in endpoint is isochronous
		
													// adr=0x0B len=1 -> always 0
		
		usbVersion = bb.getShort();					// adr=0x0C len=2	
		offsetManufacturer = bb.get();				// adr=0x0E len=1 (hard coded for 2232H=0x9A)	
		lenManufacturer = bb.get();					// adr=0x0F len=1	
		offsetProduct = bb.get();					// adr=0x10 len=1	
		lenProduct = bb.get();						// adr=0x11 len=1	
		offsetSerial = bb.get();					// adr=0x12 len=1	
		lenSerial = bb.get();						// adr=0x13 len=1
		int len = (offsetManufacturer+lenManufacturer)&0xFF;
		manufacturer = "";
		for(int i=(offsetManufacturer+2)&0xFF; i<len; i+=2)
			manufacturer += bb.getChar(i);			// adr=0x9A len=16
		len = (offsetProduct+lenProduct)&0xFF;
		product = "";
		for(int i=(offsetProduct+2)&0xFF; i<len; i+=2)
			product += bb.getChar(i);				// adr=0xAA len=32	
		len = (offsetSerial+lenSerial)&0xFF;
		serial = "";
		for(int i=(offsetSerial+2)&0xFF; i<len; i+=2)
			serial += bb.getChar(i);				// adr=0xCA len=18
		// User Eeprom Area
		bb.position(OFFSET_USER_AREA);
		boardId = bb.getShort();					// adr=0xDC len=2
		boardVersion = bb.getInt();					// adr=0xE0 len=4
		boardSerial = bb.getInt();					// adr=0xE4 len=4
		if(userDataBuffer == null)
			userDataBuffer = new byte[24];
		bb.get(userDataBuffer);
		
		// General Eeprom Area
		checksum = bb.getShort(eepromBuffer.length - 2);	// adr=0xFE len=2
		int sum = calcChecksum(eepromBuffer);
		if(checksum != sum)
			throw new IllegalArgumentException("Checksum Failure checksum=" + checksum + " != calc_sum=" + sum + ".");
	}	

	/**
	 * build the eeprom buffer. 
	 * 
	 */
	private void buildBuffer() throws IllegalArgumentException {
		
		if(eepromBuffer == null)
			eepromBuffer = new byte[256];
		
		for(int i=0; i<eepromBuffer.length; i++)
			eepromBuffer[i] = 0;
		
		ByteBuffer bb = ByteBuffer.wrap(eepromBuffer);
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.clear();

		
		portA = 8;									// set default value for SPI
		bb.put(portA);								// adr=0x00 len=1	(8->SPI, 9->async FIFO)
		portB = 9;									// set default value for async FIFO
		bb.put(portB);								// adr=0x01 len=1	(8->SPI, 9->async FIFO)
		bb.putShort(vendorId);						// adr=0x02 len=2
		bb.putShort(productId);						// adr=0x04 len=2	
		bb.putShort(chipType);						// adr=0x06 len=2	
		bb.put(configDesc);							// adr=0x08 len=1
													// 		Bit 7: always 1
													// 		Bit 6: 1 if is self powered, 0 if bus powered
													// 		Bit 5: 1 if uses remote wakeup
													// 		Bit 4: 1 if is battery powered	
		bb.put(maxPower);							// adr=0x09 len=1	
		bb.putShort(chipConfig);							// adr=0x0A len=1
													// 		Bit 4: 1 change USB version
													// 		Bit 3: 1 use the serial number string
													// 		Bit 2: 1 enable suspend pull downs
													// 		Bit 1: 1 out endpoint is isochronous
													// 		Bit 0: 1 in endpoint is isochronous
		
													// adr=0x0B len=1 -> always 0
		
		bb.putShort(usbVersion);					// adr=0x0C len=2
		offsetManufacturer = (byte) OFFSET_STRING_AREA;
		bb.put(offsetManufacturer);					// adr=0x0E len=1 (hard coded for 2232H=0x9A)
		lenManufacturer = (byte) (((manufacturer.length()*2)+2) & 0xFF);
		bb.put(lenManufacturer);	// adr=0x0F len=1
		offsetProduct = (byte) ((offsetManufacturer + lenManufacturer) & 0xFF);
		bb.put(offsetProduct);						// adr=0x10 len=1
		if(((product.length() * 2) + 2) > 32)
			lenProduct = 32;
		else
			lenProduct = (byte) (((product.length()*2)+2) & 0xFF);
		bb.put(lenProduct);							// adr=0x11 len=1	
		offsetSerial = (byte) ((offsetProduct + lenProduct) & 0xFF);
		bb.put(offsetSerial);						// adr=0x12 len=1
		serial = String.format("TBAG%04d", boardId);
		lenSerial = (byte) (((serial.length()*2)+2) & 0xFF);
		bb.put(lenSerial);							// adr=0x13 len=1
		bb.putInt(0);								// adr=0x14 len=4
		bb.put((byte) 0x56);						// adr=0x18 len=1	0x56
		
		bb.position(OFFSET_STRING_AREA);
		bb.put(lenManufacturer);
		bb.put((byte) 3);
		for(int i=0; i<manufacturer.length(); i++)
			bb.putChar(manufacturer.charAt(i));

		bb.put(lenProduct);
		bb.put((byte) 3);
		for(int i=0; i<((lenProduct / 2) - 1); i++)
			bb.putChar(product.charAt(i));

		bb.put(lenSerial);
		bb.put((byte) 3);
		for(int i=0; i<serial.length(); i++)
			bb.putChar(serial.charAt(i));

		// User Eeprom Area
		bb.position(OFFSET_USER_AREA);
		bb.putShort(boardId);						// adr=0xDC len=2
		bb.putInt(boardVersion);					// adr=0xE0 len=4
		bb.putInt(boardSerial);						// adr=0xE4 len=4
		if(userDataBuffer != null)
			bb.put(userDataBuffer);
		// General Eeprom Area
		// checksum adr=0xFE len=2
		short scheck = calcChecksum(eepromBuffer);
		bb.putShort(eepromBuffer.length - 2, scheck);				
	}

	private short calcChecksum(byte[] buffer) {
		int value = 0;
		int check = 0xAAAA;
		short ret = 0;
				
		for(int i=0; i<(buffer.length/2)-1; i++)
		{
			value = buffer[i*2] << 8;
			value += buffer[(i*2)+1];
			value &= 0xFFFF;
			check ^= value;
			check = ((check << 1) | (check>> 15)) & 0xFFFF;
		}
		ret = (short)((check & 0xFF) << 8);
		ret += (short)((check & 0xFF00) >> 8);
		return ret;	
	}
		
}
