package ch.thomson.ucs.ftdi;

import java.util.NoSuchElementException;

//import com.ftdichip.device.resource.Messages;

/**
 * Enumeration of all known errors of the FTD2xx API.
 * 
 * @author Mike Werner
 */
public enum FTStatus {
	/**
	 * No error. The operation was successful.
	 */
	FT_OK,

	/**
	 * The handle provided as a parameter is invalid.
	 */
	FT_INVALID_HANDLE,

	/**
	 * The specified device was not found.
	 */
	FT_DEVICE_NOT_FOUND,

	/**
	 * Tried to access a device that was not opened before.
	 */
	FT_DEVICE_NOT_OPENED,

	/**
	 * A general I/O error has occurred.
	 */
	FT_IO_ERROR,

	/**
	 * Not enough resources to perform the operation.
	 */
	FT_INSUFFICIENT_RESOURCES,

	/**
	 * A parameter provided to an operation was invalid.
	 */
	FT_INVALID_PARAMETER,

	/**
	 * The baud rate is invalid.
	 */
	FT_INVALID_BAUD_RATE,

	/**
	 * Tried to erase on a device that was not opened with erase permission.
	 */
	FT_DEVICE_NOT_OPENED_FOR_ERASE,

	/**
	 * Tried to write on a device that was not opened with write permission.
	 */
	FT_DEVICE_NOT_OPENED_FOR_WRITE,

	/**
	 * Writing to the device has failed.
	 */
	FT_FAILED_TO_WRITE_DEVICE,

	/**
	 * Failed to read from the EEPROM.
	 */
	FT_EEPROM_READ_FAILED,

	/**
	 * Failed to write to the EEPROM.
	 */
	FT_EEPROM_WRITE_FAILED,

	/**
	 * Failed to erase the EEPROM.
	 */
	FT_EEPROM_ERASE_FAILED,

	/**
	 * The EEPROM is not present on the device.
	 */
	FT_EEPROM_NOT_PRESENT,

	/**
	 * The EEPROM is not programmed.
	 */
	FT_EEPROM_NOT_PROGRAMMED,

	/**
	 * The arguments provided to an operation are invalid.
	 */
	FT_INVALID_ARGS,

	/**
	 * The operation is not supported for the current device.
	 */
	FT_NOT_SUPPORTED,

	/**
	 * An error has occurred that can not be described in more detail.
	 */
	FT_OTHER_ERROR;

	/**
	 * Retrieves a status value with a particular ordinal.
	 * 
	 * @param ordinal
	 *            the ordinal of the status value to search for.
	 * @return the status with the particular ordinal.
	 */
	public static FTStatus byOrdinal(int ordinal) {

		FTStatus result = null;
		for (FTStatus status : values()) {
			if (status.ordinal() == ordinal) {
				result = status;
				break;
			}
		}

		if (result == null) {
			throw new NoSuchElementException("Unknown ordinal: " + ordinal); //$NON-NLS-1$
		}

		return result;
	}

	/**
	 * Retrieves the localized message for a particular <code>FTStatus</code>.
	 * 
	 * @param status
	 *            the status value for which to retrieve the localized message.
	 * @return the localized message.
	 */
	public static String getLocalizedMessage(FTStatus status) {

/*		switch (status) {
		case FT_OK:
			return Messages.FTStatus_1;
		case FT_INVALID_HANDLE:
			return Messages.FTStatus_2;
		case FT_DEVICE_NOT_FOUND:
			return Messages.FTStatus_3;
		case FT_DEVICE_NOT_OPENED:
			return Messages.FTStatus_4;
		case FT_IO_ERROR:
			return Messages.FTStatus_5;
		case FT_INSUFFICIENT_RESOURCES:
			return Messages.FTStatus_6;
		case FT_INVALID_PARAMETER:
			return Messages.FTStatus_7;
		case FT_INVALID_BAUD_RATE:
			return Messages.FTStatus_8;
		case FT_DEVICE_NOT_OPENED_FOR_ERASE:
			return Messages.FTStatus_9;
		case FT_DEVICE_NOT_OPENED_FOR_WRITE:
			return Messages.FTStatus_10;
		case FT_FAILED_TO_WRITE_DEVICE:
			return Messages.FTStatus_11;
		case FT_EEPROM_READ_FAILED:
			return Messages.FTStatus_12;
		case FT_EEPROM_WRITE_FAILED:
			return Messages.FTStatus_13;
		case FT_EEPROM_ERASE_FAILED:
			return Messages.FTStatus_14;
		case FT_EEPROM_NOT_PRESENT:
			return Messages.FTStatus_15;
		case FT_EEPROM_NOT_PROGRAMMED:
			return Messages.FTStatus_16;
		case FT_INVALID_ARGS:
			return Messages.FTStatus_17;
		case FT_NOT_SUPPORTED:
			return Messages.FTStatus_18;
		case FT_OTHER_ERROR:
			return Messages.FTStatus_19;
		}

		return Messages.FTStatus_1;
*/		return "FTStatus";
	}

	/**
	 * Throws the exception related to a particular status value. This method is
	 * used by the native code to throw localized exceptions.
	 * 
	 * @param statusValue
	 *            the status value for which to throw the exception.
	 * @throws FTDIException
	 *             the exception related to the status value.
	 */
	static void throwException(int statusValue) throws FTDIException {
		
		String msg = "native_";
		
		if(statusValue <= -12000)
		{
			switch(statusValue / 1000){					// Master Function
			case -12:	msg += "readSpi.";							break;
			case -13:	msg += "writeSpi. ";						break;
			case -14:	msg += "writeFpga. ";						break;
			default:	msg = String.format("unknown native function: [%d]", statusValue);		break;
			}
			switch((statusValue % 1000) / 100){			// Sub Function
			case 0:		msg += "ftdispi_open: ";					break;
			case -1:	msg += "ftdispi_setmode: ";					break;
			case -3:	msg += "ftdispi_setloopback: ";				break;
			case -4:	msg += "ftdispi_write_read: ";				break;
			case -5:	msg += "ftdispi_setgpo: ";				break;
			default:	msg = String.format("unknown native function: [%d]", statusValue);		break;
			}
			switch(statusValue % 100){					// return value (fault cause)
			case -1:	msg += "[-1] context failure";				break;
			case -2:	msg += "[-2] command failed";				break;
			case -3:	msg += "[-3] memory allocation failure";	break;
			case -4:	msg += "[-4] library libftdi failure";		break;
			case -5:	msg += "[-5] timing failure";				break;
			case -6:	msg += "[-6] timeout failure";				break;
			default:	msg = String.format("unknown native function: [%d]", statusValue);		break;
			}		
		}
		else if((statusValue <= -9000) && (statusValue > -10000))
		{
			if(statusValue == -9666)
				msg += "read.ftdi_read_data: [-666] USB device unavailable";
			else
				msg += String.format("read.ftdi_read_data: [%d] usb_bulk_read() failed", statusValue % 1000);
		}
		else
		{
			switch(statusValue){			
			case -1001:		msg += "open.ftdi_init: [-1] couldn't allocate read buffer";		break;
			case -1101:		msg += "open.ftdi_set_interface: [-1] unknown interface";			break;
			case -1102:		msg += "open.ftdi_set_interface: [-2] USB device unavailable";		break;
			case -1201:		msg += "open.ftdi_usb_open: [-1] usb_find_busses() failed";			break;
			case -1202:		msg += "open.ftdi_usb_open: [-2] usb_find_devices() failed";		break;
			case -1203:		msg += "open.ftdi_usb_open: [-3] USB device not found";				break;
			case -1204:		msg += "open.ftdi_usb_open: [-4] unable to open device";			break;
			case -1205:		msg += "open.ftdi_usb_open: [-5] unable to claim device";			break;
			case -1206:		msg += "open.ftdi_usb_open: [-6] reset failed";						break;
			case -1207:		msg += "open.ftdi_usb_open: [-7] set baudrate failed";				break;
			case -1208:		msg += "open.ftdi_usb_open: [-8] get product description failed";	break;
			case -1209:		msg += "open.ftdi_usb_open: [-9] get serial number failed";			break;
			case -1210:		msg += "open.ftdi_usb_open: [-10] unable to close device";			break;
			case -1211:		msg += "open.ftdi_usb_open: [-11] ftdi context invalid";			break;
			
			case -2001:		msg += "close.ftdi_usb_close: [-1] usb_release() failed";			break;
			case -2002:		msg += "close.ftdi_usb_close: [-2] usb_close() failed";				break;
			case -2003:		msg += "close.ftdi_usb_close: [-3] ftdi context invalid";			break;
	
			case -3001:		msg += "reset.ftdi_usb_reset: [-1] FTDI reset failed";				break;
			case -3002:		msg += "reset.ftdi_usb_reset: [-2] USB device unavailable";			break;
	
			case -4001:		msg += "purgeReceiveBuffer.ftdi_usb_purge_rx_buffer: [-1] read buffer purge failed";	break;
			case -4002:		msg += "purgeReceiveBuffer.ftdi_usb_purge_rx_buffer: [-2] USB device unavailable";		break;
	
			case -5001:		msg += "purgeTransmitBuffer.ftdi_usb_purge_tx_buffer: [-1] write buffer purge failed";	break;
			case -5002:		msg += "purgeTransmitBuffer.ftdi_usb_purge_tx_buffer: [-2] USB device unavailable";		break;
	
			case -6001:		msg += "getLatencyTimer.ftdi_get_latency_timer: [-1] unable to get latency timer";		break;
			case -6002:		msg += "getLatencyTimer.ftdi_get_latency_timer: [-2] USB device unavailable";			break;
	
			case -7001:		msg += "setLatencyTimer.ftdi_set_latency_timer: [-1] latency out of range";				break;
			case -7002:		msg += "setLatencyTimer.ftdi_set_latency_timer: [-2] unable to set latency timer";		break;
			case -7003:		msg += "setLatencyTimer.ftdi_set_latency_timer: [-3] USB device unavailable";			break;
	
			case -8001:		msg += "setFlowControl.ftdi_setflowctrl: [-1] set flow control failed";					break;
			case -8002:		msg += "setFlowControl.ftdi_setflowctrl: [-2] USB device unavailable";					break;

			case -10001:	msg += "readEeprom.ftdi_read_eeprom: [-1] read failed";					break;
			case -10002:	msg += "readEeprom.ftdi_read_eeprom: [-2] USB device unavailable";		break;

			case -11001:	msg += "writeEeprom.ftdi_write_eeprom: [-1] write failed";				break;
			case -11002:	msg += "writeEeprom.ftdi_write_eeprom: [-2] USB device unavailable";	break;
			default:	msg = String.format("unknown native function: [%d]", statusValue);			break;
			}
		}
		throw new FTDIException(msg);
/*		try {
			FTStatus status = FTStatus.byOrdinal(statusValue);
			throwNewException(status, getLocalizedMessage(status));
		} catch (NoSuchElementException e) {
//			String message = String.format(Messages.FTStatus_0, statusValue);
			String message = String.format("FTStatus %d", statusValue);
			throw new FTDIException(message);
		} catch (FTDIException e) {
			throw e;
		} catch (Exception e) {
			throw (RuntimeException) e;
		}
*/		
	}

	private static void throwNewException(FTStatus status, String message)
			throws Exception {

		Exception result = null;

		switch (status) {
		case FT_INVALID_HANDLE:
			/* fall through */
		case FT_INVALID_PARAMETER:
			/* fall through */
		case FT_INVALID_BAUD_RATE:
			/* fall through */
		case FT_INVALID_ARGS:
			result = new IllegalArgumentException(message);
			break;

		case FT_DEVICE_NOT_OPENED_FOR_ERASE:
			/* fall through */
		case FT_DEVICE_NOT_OPENED_FOR_WRITE:
			/* fall through */
		case FT_DEVICE_NOT_OPENED:
			result = new IllegalStateException(message);
			break;

		case FT_EEPROM_READ_FAILED:
			/* fall through */
		case FT_EEPROM_WRITE_FAILED:
			/* fall through */
		case FT_EEPROM_ERASE_FAILED:
			/* fall through */
		case FT_EEPROM_NOT_PRESENT:
			/* fall through */
//		case FT_EEPROM_NOT_PROGRAMMED:
//			result = new EEPROMException(message);
			break;

		case FT_NOT_SUPPORTED:
			result = new UnsupportedOperationException(message);
			break;

		case FT_DEVICE_NOT_FOUND:
			/* fall through */
		case FT_IO_ERROR:
			/* fall through */
		case FT_INSUFFICIENT_RESOURCES:
			/* fall through */
		case FT_FAILED_TO_WRITE_DEVICE:
			/* fall through */
		default:
			result = new FTDIException(message);
			break;
		}

		throw result;
	}
}
