package ch.thomson.ucs.ftdi;

import java.nio.ByteBuffer;
import ch.thomson.ucs.utils.logging.IPluginLogger;


public class UsbDevice {
	public final static int FTDI_SPI = 8;			// FTDI type SPI - 			Port A
	public final static int FTDI_FIFO = 9;			// FTDI type async FIFO - 	Port B

	public final static int FPGA_ADR_COEF = 4104;		// EEPROM User Data Coefficient for FPGA
														// write User Data 24Byte to FPGA Offset = 0x00001008
	
	private final static IPluginLogger logger = Activator.getLogger();
	/**
	 * The handle to access the native API.
	 */
	private long handle = -1;
	/**
	 * The index to access the native API. (0->PortA, 1->PortB)
	 */
	private int index = 0;	// Port A
	/**
	 * The type to access the native API. (8->SPI, 9->FIFO)
	 */
	private int type = 8;	// SPI
	/**
	 * The Board ID to access the native API. (0->all Devices, 1-n->specific device)
	 */
	private int boardId = 0;	// all devices
	/**
	 * The Board Number to access the native API. (0->scalar Devices, 1-n->array index of device)
	 */
	private int boardNo = 0;	// scalar devices
	/**
	 * The Descriptor Object
	 */
	private int interfaceVer = 0;	// interface version given from the midleware
	/**
	 * The Descriptor Object
	 */
	private Descriptor descriptor = null;	// scalar devices
	/**
	 * The Firmware Object
	 */
	private Firmware firmware = null;		// Firmware Header
	/**
	 * The Buffer for FIFO read
	 */
	private byte[] fifoBuffer = null;		// FIFO Buffer
	/**
	 * The Fifo event counter to detect Drop Outs
	 */
	private int fifoEventCounter = 0;		// FIFO Event Counter
	private int fifoPulseCounter = 0;		// FIFO Event Counter
	/**
	 * The Fifo Drop Out counter to count Drop Outs
	 */
	private int dropOutEventCounter = 0;		// FIFO Drop Out Counter
	private int dropOutPulseCounter = 0;		// FIFO Drop Out Counter
	
	// time checking for fast read
//	private long lastReadFifo = 0;
//	private long startReadFifo = 0;
	
	byte[] userDataBuf = null;			// user data buffer 12 * 4 bytes written to FPGA 
			
	/**
	 * @return the firmware
	 */
	public Firmware getFirmware() {
		return firmware;
	}
	/**
	 * Creates a new <code>Device</code>. 
	 * 
	 */
	public UsbDevice() {
	}			
	/**
	 * Creates a new <code>Device</code>. 
	 * 
	 */
	public UsbDevice(int index) {
		this.index = index;			// Port B
		if((index % 2) == 1)		// jeder ungerade index ist Port B -> FIFO 
			this.type = 9;			// FIFO
	}			
	/**
	 * Creates a new <code>Device</code>. 
	 * 
	 */
	public UsbDevice(int type, int board_id, int board_no, int interfaceVer) {
		logger.info("UsbDevice( type=" + type + ", board_id=" + board_id + 
				             ", board_no=" + board_no + ", interfaceVer=" + interfaceVer + ")");
		if(type == FTDI_FIFO)
			this.index = 1;						// FIFO always on Port B
		else
			this.interfaceVer = interfaceVer; 	// only interface verification possible if type is SPI
		this.type = type;
		this.boardId = board_id;
		this.boardNo = board_no;
	}			
	
	/**
	 * Initialize this device and sets the default timeout values.
	 * 
	 * @throws FTDIException
	 *             if the device can not be opened.
	 */
	public synchronized void init() throws FTDIException {
		logger.info("UsbDevice.init() start");
		try {
			close();
			handle = -1;
		}
		catch(FTDIException e) {
			handle = -1;
			logger.info("UsbDevice.init() close() FTDIException e = " + e.getMessage());
			throw e;
		}		
		try {
			open();
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() open() FTDIException e = " + e.getMessage());
			throw e;
		}
		logger.info("open device index = " + index + " OK handle = " + handle);
		try {
			reset(handle);
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() reset(handle) FTDIException e = " + e.getMessage());
			throw e;
		}
/*		try {
			purgeReceiveBuffer(handle);
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() purgeReceiveBuffer(handle) FTDIException e = " + e.getMessage());
			throw e;
		}
		try {
			purgeTransmitBuffer(handle);
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() purgeTransmitBuffer(handle) FTDIException e = " + e.getMessage());
			throw e;
		}
*/		
		try {
			setLatencyTimer(handle, 2);
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() setLatencyTimer(handle, 2) FTDIException e = " + e.getMessage());
			throw e;
		}
		try {
			setFlowControl(handle, 0x0100, (char) 0, (char) 0);
		}
		catch(FTDIException e) {
			logger.info("UsbDevice.init() setFlowControl(handle, 0x0100, (char) 0, (char) 0) FTDIException e = " + e.getMessage());
			throw e;
		}
		if(type == FTDI_SPI) {
			byte [] buffer = new byte[256];
			try {
				this.read_eeprom(buffer);
			}
			catch(FTDIException e) {
				logger.info("UsbDevice.init() read_eeprom(buffer) FTDIException e = " + e.getMessage());
				throw e;
			}
			descriptor = new Descriptor(buffer);
						
			if(this.interfaceVer != 0)
			{
				byte [] buf = new byte[36];
				try {
					this.read(buf, 0, buf.length);
				}
				catch(FTDIException e) {
					logger.info("UsbDevice.init() read(buf, 0, buf.length) FTDIException e = " + e.getMessage());
					throw e;
				}
				firmware = new Firmware(buf);
				if((this.interfaceVer != firmware.getInterfaceVer()))
				{
					logger.info(String.format("UsbDevice.init() Different Interface Version - Middleware=%x != FPGA=%x", this.interfaceVer, firmware.getInterfaceVer()));
					throw new FTDIException(String.format("Different Interface Version - Middleware=%x != FPGA=%x", this.interfaceVer, firmware.getInterfaceVer()));
				}
			}

// MH 19.04.12	Auskommentiert in der Entwicklungsphase zum direkten schreiben der Gain and Offset				
/*			
			// write User Data 24Byte to FPGA Offset = 0x00001008
//			byte[] userDataBuf = descriptor.getUserDataBuffer().clone();
			userDataBuf = new byte[descriptor.getUserDataBuffer().length * 2];
			for(int i=0; i<userDataBuf.length; i+=4) {
				userDataBuf[i+0] = 0;
				userDataBuf[i+1] = 0;
				userDataBuf[i+2] = descriptor.getUserDataBuffer()[(i/2)+0];
				userDataBuf[i+3] = descriptor.getUserDataBuffer()[(i/2)+1];
			}
			try {
				this.write(userDataBuf, FPGA_ADR_COEF, userDataBuf.length);
			}
			catch(FTDIException e) {
				logger.info("UsbDevice.init() write(userDataBuf, 4104, userDataBuf.length) FTDIException e = " + e.getMessage());
				throw e;
			}
*/			
		}
		else {
			try {
				setLatencyTimer(handle, 1);
			}
			catch(FTDIException e) {
				logger.info("UsbDevice.init() setLatencyTimer(handle, 200) FIFO FTDIException e = " + e.getMessage());
				throw e;
			}			
		}			
	}

	/**
	 * Get the Eeprom Buffer as parsed Descriptor.
	 * 
	 * @throws FTDIException
	 *             if the Descriptor not exist.
	 */
	public Descriptor getDescriptor() throws IllegalArgumentException {
		if (descriptor == null)
			throw new IllegalArgumentException("The Descriptor object not exist.");
		return descriptor;
	}
	
	/**
	 * Set the Eeprom Buffer as build Descriptor.
	 * 
	 * @throws FTDIException
	 *             if the Descriptor not exist.
	 */
	public void setDescriptor(Descriptor desc) throws IllegalArgumentException, FTDIException {
		if (desc == null)
			throw new IllegalArgumentException("The Descriptor object not exist.");
		this.write_eeprom(desc.getEepromBuffer());
		descriptor = desc;
	}

	/**
	 * Set the Eeprom Buffer as build Descriptor.
	 * 
	 * @throws FTDIException
	 *             if the Descriptor not exist.
	 */
	public void setDescriptor(short id, int version, int serial) throws IllegalArgumentException, FTDIException {
		if(descriptor == null) {
			descriptor = new Descriptor();
		}
		descriptor.setUABoardId(id);	
		descriptor.setUABoardVersion(version);	
		descriptor.setUABoardSerial(serial);	
		this.write_eeprom(descriptor.getEepromBuffer());
	}

	/**
	 * Set the Eeprom Buffer as build Descriptor.
	 * 
	 * @throws FTDIException
	 *             if the Descriptor not exist.
	 */
	public void setDescriptor(String product, short id, int version, int serial) throws IllegalArgumentException, FTDIException {
		if(descriptor == null) {
			descriptor = new Descriptor();
		}
		descriptor.setProduct(product);
		setDescriptor(id, version, serial);
	}

	/**
	 * Set the Eeprom Buffer as build Descriptor.
	 * 
	 * @throws FTDIException
	 *             if the Descriptor not exist.
	 */
	public void setDescriptor(boolean selfpow, String product, short id, int version, int serial) throws IllegalArgumentException, FTDIException {
		if(descriptor == null) {
			descriptor = new Descriptor();
		}
		descriptor.setSelfPowered(selfpow);
		setDescriptor(product, id, version, serial);
	}

	/**
	 * open this device.
	 * 
	 * @throws FTDIException
	 *             if the device can not be closed.
	 */
	public void open() throws FTDIException {
		if(boardId == 0)
		{
			logger.info("UsbDevice.open(" + index + ", \"\")");
			handle = open(index, "");
		}
		else
		{
			String serial = String.format("TBAG%04d", boardId);
			logger.info("UsbDevice.open(" + index + "," + serial + ")");
			handle = open(index, serial);
		}
	}

	/**
	 * close this device.
	 * 
	 * @throws FTDIException
	 *             if the device can not be closed.
	 */
	public void close() throws FTDIException {
		if(handle != -1)
		{
			close(handle);
		}
	}
	
	/**
	 * Reads reads a specified number of bytes. If no data is available, it
	 * blocks until one was received or the end of the stream was detected.
	 * 
	 * @param buffer
	 *            the buffer to store the data in.
	 * @param offset
	 *            the index within the <code>buffer</code> of the first byte
	 *            to store.
	 * @param length
	 *            the number of bytes to read.
	 * @return the number of bytes actually read, or -1 if the end of the stream
	 *         was detected.
	 * @throws FTDIException
	 *             if the operation can not be performed.
	 */
	public int read(byte[] buffer, int offset, int length)
			throws FTDIException {

		int result = -1; 
			
		if (buffer == null || buffer.length == 0)
			throw new IllegalArgumentException("The buffer is empty.");

		try{
			if(handle == -1)
				init();
			if(handle != -1)
			{
				if(type == 8)	// SPI
					result = readSpi(handle, buffer, offset, length);
				else
					result = readFifo(buffer);
			}
		}
		catch (FTDIException e) {
			logger.info("readSpi / readFifo type=" + type + " FTDIException e=" + e.getMessage());
			init();
			throw e;
		}
		return result;
	}
		
	private int readFifo(byte[] buffer) throws FTDIException {		
		// Input Buffer
		if(fifoBuffer == null)
			fifoBuffer = new byte[16384];
		// Output Buffer
		if(buffer.length < 16010)
			throw new FTDIException(String.format("readFifo: Buffer to small %d < 16010", buffer.length));
		
//		startReadFifo = System.currentTimeMillis();
		int result = read(handle, fifoBuffer, 0, fifoBuffer.length);
		
		if(result == 0)
			return 0; // Es gibt nichts zu lesen, da keine Daten vorhanden
		
		result = parseFifoBuffer(fifoBuffer, buffer);
		// Test
/*		while(result == -1)
		{
			if(smallBuf == null)
				smallBuf = new byte[4096];
			result = read(handle, smallBuf, 0, smallBuf.length);
			ByteBuffer bbIn = ByteBuffer.wrap(smallBuf);
			bbIn.order(ByteOrder.LITTLE_ENDIAN);
			int start = bbIn.getShort();
			if(start == 0x8001)
			{
				if(bigBuf == null)
					bigBuf = new byte[fifoBuffer.length - smallBuf.length];
				result = read(handle, bigBuf, 0, bigBuf.length);
				ByteBuffer bb = ByteBuffer.wrap(fifoBuffer);
				bb.put(smallBuf);
				bb.put(bigBuf);
				result = parseFifoBuffer(bb.array(), buffer);
			}
			else
			{
				logger.info(String.format("[Board ID: %d] parseFifoBuffer: Start sign %x != 0x8001", boardId, start));
				result = -1;
			}
		}
*/
//		lastReadFifo = System.currentTimeMillis();
		return result;
	}
	
	private int parseFifoBuffer(byte[] fifoBuf, byte[] buffer) throws FTDIException {
		// TODO Auto-generated method stub
		ByteBuffer bbIn = ByteBuffer.wrap(fifoBuf);
//		bbIn.order(ByteOrder.LITTLE_ENDIAN);
		ByteBuffer bbOut = ByteBuffer.wrap(buffer);
		
		
		short start = bbIn.getShort();
		if((start & 0x8001) != 0x8001)
		{
			logger.error(String.format("[Board ID: %d] parseFifoBuffer: Start sign 0x%x != 0x8001 (buf[0]=0x%x, buf[1]=0x%x)", boardId, start, fifoBuf[0], fifoBuf[1]));
			return -1;
		}
		int eventCounter = bbIn.getInt();
		if((eventCounter != 0) && (fifoEventCounter != 0) && 
		   ((fifoEventCounter + 1) != eventCounter) && (fifoEventCounter != eventCounter))
		{
			dropOutEventCounter++;
			logger.error(String.format("[Board ID: %d] FIFO Sync Drop Out count=%d (old event count = %d new event count = %d)", boardId, dropOutEventCounter, fifoEventCounter, eventCounter));
//			long act_time = System.currentTimeMillis();			
//			logger.error(String.format("[Board ID: %d] FIFO Sync Drop Out TimeLast=%dms TimeRead=%dms", boardId, act_time - lastReadFifo, act_time - startReadFifo));
		}
		fifoEventCounter = eventCounter;

		int pulseCounter = bbIn.getInt();
		if((pulseCounter != 0) && (fifoPulseCounter != 0) && 
		   ((fifoPulseCounter + 1) != pulseCounter) && (fifoPulseCounter != pulseCounter))
		{
			dropOutPulseCounter++;
			logger.error(String.format("[Board ID: %d] FIFO Pulse Drop Out count=%d (old event count = %d new event count = %d)", boardId, dropOutPulseCounter, fifoPulseCounter, pulseCounter));
//			long act_time = System.currentTimeMillis();			
//			logger.error(String.format("[Board ID: %d] FIFO Pulse Drop Out TimeLast=%dms TimeRead=%dms", boardId, act_time - lastReadFifo, act_time - startReadFifo));
		}
		fifoPulseCounter = pulseCounter;
		
		bbOut.putInt(eventCounter);			// Event Counter 32Bit 
		bbOut.putInt(pulseCounter);			// Pulse Counter 32bit
		bbOut.putShort(bbIn.getShort());	// Reference Voltage 16bit
		int offset = bbOut.position(); 		// Bytes before Measurement block
		
		// TODO: Schleife ersetzen durch native array copy befehle... 
		for(int i=0; i<2000; i++){
			bbOut.putShort(offset + (i*2), 			bbIn.getShort());
			bbOut.putShort(offset + 4000 + (i*2), 	bbIn.getShort());
			bbOut.putShort(offset + 8000 + (i*2), 	bbIn.getShort());
			bbOut.putShort(offset + 12000 + (i*2), 	bbIn.getShort());
		}
		short stop = bbIn.getShort(16012);
		if((stop & 0xC001) != 0xC001)
			throw new FTDIException(String.format("parseFifoBuffer: Stop sign 0x%x != 0xC001 (buf[0]=0x%x, buf[1]=0x%x)", stop, fifoBuf[16012], fifoBuf[16013]));
		return 1;
	}

	public int write(byte[] buffer, int offset, int length)
			throws FTDIException {
		
		int result = -1; 
		
		if (buffer == null || buffer.length == 0)
			throw new IllegalArgumentException("The buffer is empty.");

		if(type == 8)	// SPI only support write function
		{
			try{
// MH 19.04.12	Auskommentiert in der Entwicklungsphase zum direkten schreiben der Gain and Offset				
/*				if(offset == (FPGA_ADR_COEF - 8)) { 
					if(userDataBuf == null)
						throw new IllegalArgumentException("The user data buffer not exist.");
						
					if(buffer.length < (userDataBuf.length + 8))
						throw new IllegalArgumentException("The buffer length=" + buffer.length + 
								" is to small for user data=" + userDataBuf.length + "+8.");
					
					for(int i=0; i<userDataBuf.length; i++)
						buffer[i+8] = userDataBuf[i];
				}
*/				
				if(handle == -1)
					init();
				if(handle != -1)
					result = writeSpi(handle, buffer, offset, length);
			}
			catch (FTDIException e) {
				logger.info("writeSpi FTDIException e=" + e.getMessage());
				init();		// close connection set handle = -1
				throw e;
			}
		}
		return result;
	}
	
	public int writeFpga(byte[] buffer, int offset, int length)
		throws FTDIException, InterruptedException {
	
		if (buffer == null || buffer.length == 0)
			throw new IllegalArgumentException("The buffer is empty.");
		
		if (offset < 0 || (offset + length) > buffer.length)
			throw new IllegalArgumentException("The buffer is to small.");
		
		if (type != 8)
			throw new IllegalArgumentException("Wrong Type SPI only supported.");
		
		// write firmware
		writeFpga(handle, buffer, offset, length);
		byte buf[] = {1};
		// set Remote Update Control Register
		writeSpi(handle, buf, 0x1000, buf.length);
		// set Reconfig Flag
		writeSpi(handle, buf, 0x1004, buf.length);
		long start_time = System.currentTimeMillis();
		do {
			if((readSpi(handle, buf, 0x32, buf.length) == buf.length) && (buf[0] == 1))
			{
				buf[0] = 0;
				// set Remote Update Control Register
				writeSpi(handle, buf, 0x1000, buf.length);
				// set Reconfig Flag
				writeSpi(handle, buf, 0x1004, buf.length);
				return 0;
			}
			Thread.sleep(50);
		}while((start_time + 10000) > System.currentTimeMillis());

		return -1;	// timeout 
	}

	public int read_eeprom(byte[] buffer) throws FTDIException {

		int result = -1; 
		
		if (buffer == null)
			throw new IllegalArgumentException("The buffer is empty.");

		if (buffer.length != 256)
			throw new IllegalArgumentException("The buffer is not 256 byte long.");

		try{
			result = readEeprom(handle, buffer, buffer.length);
		}
		catch (FTDIException e) {
			logger.info("readEeprom FTDIException e=" + e.getMessage());
			try{
				init();
				result = readEeprom(handle, buffer, buffer.length);
			}
			catch (FTDIException e1) {
				logger.error("init FTDIException e1=" + e1.getMessage());
				throw e1;
			}
		}
		return result;
	}	
	
	public int write_eeprom(byte[] buffer) throws FTDIException {

		int result = -1; 

		if (buffer == null)
			throw new IllegalArgumentException("The buffer is empty.");

		if (buffer.length != 256)
			throw new IllegalArgumentException("The buffer is not 256 byte long.");

		try{
			result = writeEeprom(handle, buffer, buffer.length);
		}
		catch (FTDIException e) {
			logger.info("writeEeprom FTDIException e=" + e.getMessage());
			try{
				init();
				result = writeEeprom(handle, buffer, buffer.length);
			}
			catch (FTDIException e1) {
				throw e1;
			}
		}
		return result;
	}	

	/**
	 * Opens a device by it`s serial number.
	 * 
	 * @param locationID
	 *            the device`s location within the USB tree.
	 * @return the native handle of the opened device.
	 * @throws FTDIException
	 *             if the device can not be opened.
	 */
	private static native long open(int index, String serial) throws FTDIException;

	/**
	 * Closes a device.
	 * 
	 * @param handle
	 *            the native handle of the device to close.
	 * @throws FTDIException
	 *             if the device can not be closed.
	 */
	private native static void close(long handle) throws FTDIException;
	
	/**
	 * Resets this device.
	 * 
	 * @throws FTDIException
	 *             if the device can not be reset.
	 */
	private static native void reset(long handle) throws FTDIException;
	
	/**
	 * Purges the receive buffer of a particular device.
	 * 
	 * @throws FTDIException
	 *             if the device can not be accessed.
	 */
	private static native void purgeReceiveBuffer(long handle)
			throws FTDIException;

	/**
	 * Purges the transmit buffer of a particular device.
	 * 
	 * @throws FTDIException
	 *             if the device can not be accessed.
	 */
	private static native void purgeTransmitBuffer(long handle)
			throws FTDIException;
	
	/**
	 * Retrieves the current value of the latency timer. <br>
	 * <br>
	 * <b>Note: </b> this method is only valid for FT232BM and FT245BM devices.
	 * 
	 * @param handle
	 *            the native handle of the device for which to retrieve the
	 *            timer value.
	 * @return the latency timer value.
	 * @throws FTDIException
	 *             if the value can not be retrieved.
	 */
	private static native int getLatencyTimer(long handle) throws FTDIException;

	/**
	 * Sets the new latency timer value of a particular device.
	 * 
	 * @param handle
	 *            the native handle of the device.
	 * @param value
	 *            the new value in milliseconds. The value range is 1 - 255.
	 * @throws FTDIException
	 *             if the value can not be set.
	 */
	private static native void setLatencyTimer(long handle, int value) throws FTDIException;

	/**
	 * Sets the flow control mode of a particular device`s port.
	 * 
	 * @param handle
	 *            the native handle of the device.
	 * @param flowControl
	 *            the new flow control mode.
	 * @param xon
	 *            the character to signal XON (only used for XON XOFF mode).
	 * @param xoff
	 *            the character to signal XOFF (only used for XON XOFF mode).
	 * @throws FTDIException
	 *             if the flow control mode can not be set.
	 */
	private static native void setFlowControl(long handle, int flowControl,
			char xon, char xoff) throws FTDIException;

	/**
	 * Reads reads a specified number of bytes from a particular device. If no
	 * data is available, it blocks until one was received or the end of the
	 * stream was detected.
	 * 
	 * @param handle
	 *            the native device handle.
	 * @param buffer
	 *            the buffer to store the data in.
	 * @param offset
	 *            the index within the <code>buffer</code> of the first byte
	 *            to store.
	 * @param length
	 *            the number of bytes to read.
	 * @return the number of bytes actually read, or -1 if the end of the stream
	 *         was detected.
	 * @throws FTDIException
	 *             if the operation can not be performed.
	 */
	private static native int read(long handle, byte[] buffer,
			int offset, int length) throws FTDIException;
	
	private synchronized static native int readSpi(long handle, byte[] buffer,
			int offset, int length) throws FTDIException;

	private synchronized static native int writeSpi(long handle, byte[] buffer,
			int offset, int length) throws FTDIException;
	
	private synchronized static native int writeFpga(long handle, byte[] buffer,
			int offset, int length) throws FTDIException;

	/**
	 * Reads reads a specified number of bytes from a particular device. If no
	 * data is available, it blocks until one was received or the end of the
	 * stream was detected.
	 * 
	 * @param handle
	 *            the native device handle.
	 * @param buffer
	 *            the buffer to store the data in.
	 * @param offset
	 *            the index within the <code>buffer</code> of the first byte
	 *            to store.
	 * @param length
	 *            the number of bytes to read.
	 * @return the number of bytes actually read, or -1 if the end of the stream
	 *         was detected.
	 * @throws FTDIException
	 *             if the operation can not be performed.
	 */
	private synchronized static native int readEeprom(long handle, byte[] buffer, int length) 
		throws FTDIException;
	
	/**
	 * Write a specified number of bytes from a particular device. If no
	 * data is available, it blocks until one was received or the end of the
	 * stream was detected.
	 * 
	 * @param handle
	 *            the native device handle.
	 * @param buffer
	 *            the buffer to store the data in.
	 * @param offset
	 *            the index within the <code>buffer</code> of the first byte
	 *            to store.
	 * @param length
	 *            the number of bytes to read.
	 * @return the number of bytes actually read, or -1 if the end of the stream
	 *         was detected.
	 * @throws FTDIException
	 *             if the operation can not be performed.
	 */
	private synchronized static native int writeEeprom(long handle, byte[] buffer, int length) 
		throws FTDIException;
}

