package ch.thomson.ucs.security;

import java.sql.SQLIntegrityConstraintViolationException;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.CopyOnWriteArrayList;

import ch.thomson.ucs.security.dao.RoleDAO;
import ch.thomson.ucs.security.dao.UserDAO;
import ch.thomson.ucs.security.dao.entries.Role;
import ch.thomson.ucs.security.dao.entries.User;
import ch.thomson.ucs.security.dao.impl.RoleDAOImpl;
import ch.thomson.ucs.security.dao.impl.UserDAOImpl;
import ch.thomson.ucs.security.helpers.BeanObject;
import ch.thomson.ucs.security.helpers.Encryptor;

public class LogonManager extends BeanObject {
	
	private static LogonManager instance;
	private final int DEFAULT_LOGON_LEVEL = 2;
	private String hostDB;
	private List<LogonLevelListener> listeners = new CopyOnWriteArrayList<LogonLevelListener>();
	
	private User loggedInUser = new User();
	private Timer timer;
	private List<User> users = Collections.emptyList();
	private List<Role> roles = Collections.emptyList();
	
	private LogonManager() {
		setLoggedInUser(null);
	}
	
	public boolean isUserLoggedIn() {
		return getLogonLevel() > DEFAULT_LOGON_LEVEL;
	}
		
	public List<User> getUsers() {
		return users;
	}

	public void setUsers(List<User> users) {
		firePropertyChange("users", this.users, this.users = users);
	}
	
	public List<Role> getRoles() {
		return roles;
	}

	public void setRoles(List<Role> roles) {
		firePropertyChange("roles", this.roles, this.roles = roles);
	}

	private void setLoggedInUser(User user) {
		if(user == null)
			user = new User();
		
		firePropertyChange("loggedInUser", this.loggedInUser, this.loggedInUser = user);
	}
	
	public User getLoggedInUser() {
		return loggedInUser;
	}
	
	public void setHostDB(String hostDB) {
		this.hostDB = hostDB;
	}
	
	public void changePassword(String oldPassword, String newPassword) throws Exception {
		if(!isUserLoggedIn())
			throw new Exception("No user logged in");
		
		if(!loggedInUser.getPassword().equals(Encryptor.toEncryptedString(oldPassword)))
			throw new Exception("Wrong password");
		
		loggedInUser.setPasswordUnencrypted(newPassword);
		this.addUser(loggedInUser);
	}
	
	public void deleteUser(String userName) throws Exception {
		// Kein User eingeloggt -> abbrechen
		if(!isUserLoggedIn())
			throw new Exception("No user is logged in");
		
		// Eingeloggter User kann sich selbst nicht lschen
		if(this.loggedInUser.getName().equals(userName))
			throw new Exception("Delete yourself is not allowed");
		
		// Es drfen nur Users gelscht werden, die einen tieferen oder gleichen Logon Level haben
		User user = getUser(userName);
		if(this.loggedInUser.getRole().compareTo(user.getRole()) > 0)
			throw new Exception("Insufficient access rights");
		
		UserDAO userDAO = createUserDAO();
		
		try {
			userDAO.open();
			userDAO.deleteUser(userName);
			this.refreshUsers();
		} catch(Exception e) {
			e.printStackTrace();
			throw new Exception("DB Connection lost");
		} finally {
			userDAO.close();
		}	
	}
	
	public void createUser(User user) throws Exception {
		// Kein User eingeloggt -> abbrechen
		if(!isUserLoggedIn())
			throw new Exception("No user is logged in");
		
		if(this.loggedInUser.getRole().compareTo(user.getRole()) > 0)
			throw new Exception("Insufficient access rights");
		
		UserDAO userDAO = this.createUserDAO();
		
		try {
			userDAO.open();
			userDAO.createUser(user);
			this.refreshUsers();
		} catch(SQLIntegrityConstraintViolationException e) {
			throw new Exception ("The user already exists");	
		} finally {
			userDAO.close();
		}		
	}
	
	private User getUser(String name) throws Exception {
		User user;
		UserDAO userDAO = createUserDAO();
		
		try {		
			userDAO.open();
			user = userDAO.getUser(name);
			
		} catch(Exception e) {
			e.printStackTrace();
			throw new Exception("DB Connection lost");
		
		} finally {
			userDAO.close();
		}
		
		if(user == null)
			throw new Exception("Unknown user");
		
		return user;
	}
	
	private void addUser(User user) throws Exception {
		UserDAO userDAO = createUserDAO();
		
		try {
			userDAO.open();
			userDAO.addUser(user);
		} catch(Exception e) {
			e.printStackTrace();
			throw new Exception("DB Connection lost");
			
		} finally {
			userDAO.close();
		}
	}

	public void logOn(String name, String password) throws Exception {
		User user = this.getUser(name);
		
		if(!user.getPassword().equals(Encryptor.toEncryptedString(password)))
			throw new Exception("Wrong password");
		
		if(isUserLoggedIn())
			logOff();
		
		setLoggedInUser(user);
		this.refreshUsers();
		this.refreshRoles();
		
		setupAutoLogOff(loggedInUser);
		firePropertyChange("userLoggedIn", false, true);
		notifyLogonLevelListeners();
	}
	
	private void setupAutoLogOff(User user) {
		// AutoLogOff erwnscht?
		if(user.getRole().getLogoffTime() == 0)
			return;
		
		// AutoLogOff Timer erzeugen
		timer = new Timer(true);
		timer.schedule(new TimerTask() {

			@Override
			public void run() {
				logOff();
			}
		}, user.getRole().getLogoffTime() * 60000);
	}
	
	private void refreshRoles() throws Exception {
		RoleDAO roleDAO = this.createRoleDAO();
		
		try {	
			roleDAO.open();	
			List<Role> list = (List<Role>) roleDAO.getAllEntries();
			
			// Liste filtern nach Logon Level
			Iterator<Role> it = list.iterator();
			while(it.hasNext()) {
				Role role = it.next();
				if(role.getLogonLevel() > this.getLogonLevel() || role.getLogonLevel() <= DEFAULT_LOGON_LEVEL)
					it.remove();
			}
			
			setRoles(list);
			
		} finally {
			roleDAO.close();
		}
	}
	
	private void refreshUsers() throws Exception {
		UserDAO userDAO = this.createUserDAO();
		
		try {
			if(userDAO.isClosed())
				userDAO.open();	
			
			List<User> list = (List<User>) userDAO.getAllEntries();
			
			// Liste filtern nach Logon Level
			Iterator<User> it = list.iterator();
			while(it.hasNext()) {
				User user = it.next();
				if(user.getRole().getLogonLevel() > getLogonLevel())
					it.remove();
			}
			
			setUsers(list);
		} finally {
			userDAO.close();
		}
	}
	
	public void logOff() {
		setLoggedInUser(null);
		List<User> users = Collections.emptyList();
		setUsers(users);
		
		List<Role> roles = Collections.emptyList();
		setRoles(roles);
		
		firePropertyChange("userLoggedIn", true, false);
		notifyLogonLevelListeners();
	}
	
	public int getLogonLevel() {
		return loggedInUser.getRole().getLogonLevel() < DEFAULT_LOGON_LEVEL ? DEFAULT_LOGON_LEVEL : loggedInUser.getRole().getLogonLevel();
	}
	
	private void notifyLogonLevelListeners() {
		Iterator<LogonLevelListener> it = listeners.iterator();
		while(it.hasNext())
			it.next().logonLevelChanged(getLogonLevel());
	}
	
	public static synchronized LogonManager getInstance() {
		if(instance == null)
			instance = new LogonManager();
		return instance;
	}
	
	public boolean addLogonLevelListener(LogonLevelListener l) {
		return listeners.add(l);
	}
	
	public boolean removeLogonLevelListener(LogonLevelListener l) {
		return listeners.remove(l);
	}
	
	private boolean tablesPrepared;
	
	private void prepareTables() throws Exception {
		if (!tablesPrepared) {
			tablesPrepared = true;
			
			RoleDAO roleDAO = this.createRoleDAO();
			UserDAO userDAO = this.createUserDAO();

			try {
				roleDAO.open();
				roleDAO.setupTable();
				userDAO.open();
				userDAO.setupTable();

			} catch(Exception e) {
				tablesPrepared = false;
				throw e;
			}
			finally {
				roleDAO.close();
				userDAO.close();
			}
		}
	}
	
	private UserDAO createUserDAO() throws Exception {
		if(!tablesPrepared)
			this.prepareTables();
		
		return new UserDAOImpl() {

			@Override
			public void open() throws Exception {
				if(hostDB == null)
					super.open();
				else
					this.open(hostDB);	
			}	
		};
	}
	
	private RoleDAO createRoleDAO() throws Exception {
		if(!tablesPrepared)
			this.prepareTables();
		
		return new RoleDAOImpl() {

			@Override
			public void open() throws Exception {
				if(hostDB == null)
					super.open();
				else
					this.open(hostDB);	
			}
		};
	}
	
	public boolean isPasswordComplexityOk(String password) {
		return password.length() >= 6;
	}
	
	public String getPasswordComplexityMessage() {
		return "Password must be at least 6 characters.";
	}
}
