package ch.thomson.swing.graphics;

import java.awt.Color;

/**
 * Eine rechteckige Skala mit Beschriftungen auf der Seite
 * <p>
 * Auf einer Skale werden in den meisten Fllen 
 * einzelne Werte als Striche oder Ticks dargestellt:<br>
 * 
 * <code>												
 * MaxValue                                  MinValue<br>
 *    +&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;+ <br>
 *    +&nbsp;.&nbsp;&nbsp;.&nbsp;&nbsp;.&nbsp;&nbsp;.&nbsp;&nbsp;.&nbsp;&nbsp;.&nbsp;&nbsp;.&nbsp;+<br>
 *    1&nbsp;&nbsp;&nbsp;3&nbsp;&nbsp;&nbsp;5&nbsp;&nbsp;&nbsp;7&nbsp;&nbsp;&nbsp;9&nbsp;&nbsp;&nbsp;11&nbsp;&nbsp;13&nbsp;&nbsp;15<br>
 * </code>
 * 
 * Zustzlich werden meisten zwei verschieden Unterteilungen
 * gemacht, eine grobe Auflsung und eine eine feine Auflsung.
 * Diesem Umstand wird in diesem interface ebefalls Rechnung getragen
 * indem verschieden folgende Optionen eingestellt werden knnen
 * 
 * <ul>
 * <li> Die Anzahl der Skalen mit <em>Interval</em></li>
 * <li> Die Anzahl der Skalenunterteilung (also Ticks) pro <em>Interval</em> mit <em>Ticks</em></li>
 * <li> Die Lnge der Ticks pro<em>Interval</em> mit <em>TicksLength</em></li>
 * <li> Die Position der Ticks pro<em>Interval</em> mit <em>TicksPosition</em></li>
 * <li> Die Farbe der Ticks pro<em>Interval</em> mit <em>TicksColor</em></li>
 * <li> Die Strichdicke in Pixel der Ticks pro<em>Interval</em> mit <em>TicksSize</em></li>
 * </ul> 
 * 
 * Um die Anwendung dieser Optionen zu verdeutlichen hier ein einfaches Beispiel:
 * 
<code>
....
import ch.thomson.webinterface.thgui.thicons.JLabel.JLabelScaleRectangular;<br>
.....<br>
.... (im Code)...<br>
  
 		int[] ticks = { 65, 17, 9, 5 };<br>
		Color[] ticksColor = { Color.green, Color.gray, Color.gray, Color.black };<br>
		int[] ticksLength = { 10, 18, 21, 30 };<br>
		int[] ticksPosition = { SwingConstants.CENTER, SwingConstants.CENTER,<br>
				SwingConstants.CENTER, SwingConstants.CENTER };<br>
		int[] ticksSize = { 1, 2, 3, 5 };<br>
		labelScaleRectangular.setSizeLine(6);<br>
		labelScaleRectangular.setInterval(4);<br>
		labelScaleRectangular.setTicks(ticks);<br>
		labelScaleRectangular.setTicksColor(ticksColor);<br>
		labelScaleRectangular.setTicksLength(ticksLength);<br>
		labelScaleRectangular.setTicksPosition(ticksPosition);<br>
		labelScaleRectangular.setTicksSize(ticksSize);<br>
<br>
 ......<br>
</code>
 * 
 * In diesem Beispiel werden 4 Skalendargestellt, jeweils mit unterschiedlicher Anzahl
 * Ticks, TicksColor und dergleichen.
 *  
 * 
 * @author Dominic Meier
 * @version $version$ ($date$)
 */
public interface IGraphicsScaleRectangular extends IGraphicsLine, IGraphicsBase{

	/**
	 * Gibt an ob die Achse angezeigt werden soll
	 * 
	 * @return <b>true</b> wenn angezeigt
	 */
	public abstract boolean isVisibleAxis();

	/**
	 * Der Wert muss auf <b>true</b> gesetzt werden um die Achse sichtbar 
	 * zu machen
	 * 
	 * @param <b>true</b> wenn anzuzeigen
	 */
	public abstract void setVisibleAxis(boolean visibleAxis);

	/**
	 * @return the interval
	 */
	public abstract int getInterval();

	/**
	 * @param interval
	 *            the interval to set
	 */
	public abstract void setInterval(int interval);

	/**
	 * @return the ticks
	 */
	public abstract int[] getTicks();

	public abstract int getTicks(int index);

	/**
	 * @param ticks
	 *            the ticks to set
	 */
	public abstract void setTicks(int[] ticks);

	public abstract void setTicks(int index, int value);

	/**
	 * @return the ticksColor
	 */
	public abstract Color[] getTicksColor();

	public abstract Color getTicksColor(int index);

	/**
	 * @param ticksColor
	 *            the ticksColor to set
	 */
	public abstract void setTicksColor(Color[] ticksColor);

	public abstract void setTicksColor(int index, Color value);

	/**
	 * @return the ticksSize
	 */
	public abstract int[] getTicksSize();

	public abstract int getTicksSize(int index);

	/**
	 * @param ticksSize
	 *            the ticksSize to set
	 */
	public abstract void setTicksSize(int[] ticksSize);

	public abstract void setTicksSize(int index, int value);

	/**
	 * @return the ticksLength
	 */
	public abstract int[] getTicksLength();

	public abstract int getTicksLength(int index);

	/**
	 * @param ticksLength
	 *            the ticksLength to set
	 */
	public abstract void setTicksLength(int[] ticksLength);

	public abstract void setTicksLength(int index, int value);

	/**
	 * @return the ticksPosition
	 */
	public abstract int[] getTicksPosition();

	public abstract int getTicksPosition(int index);

	/**
	 * @param ticksPosition
	 *            the ticksPosition to set
	 */
	public abstract void setTicksPosition(int[] ticksPosition);

	public abstract void setTicksPosition(int index, int value);

	/**
	 * Korrigiert die Lnge damit auch an den Rndern die gesammte Strichdicke
	 * sichtbar bleibt. 
	 * 
	 * @return the correctLength
	 */
	public abstract boolean isCorrectLength();

	/**
	 * @param correctLength
	 *            the correctLength to set
	 */
	public abstract void setCorrectLength(boolean correctLength);

}