package ch.thomson.ucs.tine.server.driver.modbus;

import java.net.InetAddress;
import java.nio.ByteBuffer;

import net.wimpi.modbus.ModbusIOException;
import net.wimpi.modbus.io.ModbusTCPTransaction;
import net.wimpi.modbus.msg.ReadInputDiscretesRequest;
import net.wimpi.modbus.msg.ReadInputDiscretesResponse;
import net.wimpi.modbus.msg.ReadInputRegistersRequest;
import net.wimpi.modbus.msg.ReadInputRegistersResponse;
import net.wimpi.modbus.msg.WriteMultipleCoilsRequest;
import net.wimpi.modbus.msg.WriteMultipleCoilsResponse;
import net.wimpi.modbus.msg.WriteMultipleRegistersRequest;
import net.wimpi.modbus.msg.WriteMultipleRegistersResponse;
import net.wimpi.modbus.net.TCPMasterConnection;
import net.wimpi.modbus.procimg.InputRegister;
import net.wimpi.modbus.procimg.Register;
import net.wimpi.modbus.procimg.SimpleRegister;
import net.wimpi.modbus.util.BitVector;

import org.eclipse.emf.common.util.URI;

import ch.thomson.ucs.tine.server.DataTable;
import ch.thomson.ucs.tine.server.PropertyInt32;
import ch.thomson.ucs.tine.server.RealProperty;
import ch.thomson.ucs.tine.server.impl.DriverImpl;

public class ModbusDriver extends DriverImpl {

	private URI uri;
	private TCPMasterConnection connection;
	private int unitID = 1; // Default ID
	private ModbusTCPTransaction modbusTCPTransaction;
	
	@Override
	public void initialize() throws Exception {
		// modbus://host[:port]/[type]/[unitID]
		// type = default oder wago
		
		uri = URI.createURI(getAddress());
		
		// UnitID setzen
		if(uri.segmentCount() >= 2)
			unitID = Integer.parseInt(uri.segment(1));
			
		this.getConnection();
	}
	
	/**
	 * Damit der Transactionzhler hochzhlen kann, soll immer dieselbe
	 * ModbusTCPTransaction Instanz benutzt werden (beim Neuanlegen wrde der
	 * Zhler stets auf 0 gesetzt). Diese Instanz soll treiberweit vewendet
	 * werden.
	 * 
	 * @return Die TCP-Transaction-Instanz mit der akutellen Connection
	 * @throws Exception
	 */
	private ModbusTCPTransaction getModbusTCPTransaction() throws Exception {
		
		if(this.modbusTCPTransaction == null) {
			this.modbusTCPTransaction = new ModbusTCPTransaction();
		}
		
		this.modbusTCPTransaction.setConnection(getConnection());
		return this.modbusTCPTransaction;
	}
	
	@Override
	public void writeData(DataTable dataTable) throws Exception {
		for (RealProperty property : dataTable.getProperties()) {
			if(!property.isDirty())
				continue;
			
			AddressParser addressParser = (AddressParser) property
					.getUserDefinedAddressObject();
			
			ByteBuffer bb = dataTable.getByteBuffer();
			bb.position(property.getAddressOffset());
			int values[] = new int[property.getLength()];
			
			for(int i=0; i<property.getLength(); ++i) {
				
				if(property instanceof PropertyInt32) 
					values[i] = bb.getInt();
				else
					values[i] = bb.getShort();
			}

			if (addressParser.isDiscrete()) {
				// Maximal 8 Coils knnen momentan in einer TINE-Variable bedient werden
				BitVector bv = new BitVector(addressParser.getLength() > 8 ? 8 : addressParser.getLength());
				bv.getBytes()[0] = (byte) values[0];			
				writeMultipleEntriesDiscrete(addressParser.getAddress(), bv);
			}

			else {
				Register[] regs = new Register[values.length];
				for(int i=0; i<values.length; ++i) {
					regs[i] = new SimpleRegister(values[i]);
				}
				
				writeMultipleEntries(addressParser.getAddress(), regs);
			}
		}
	}
	
	private synchronized TCPMasterConnection getConnection() throws Exception {
		if(connection != null) {
			if(!connection.isConnected()) {
				connection.connect();
			}
		}
		
		if(connection == null) {
			connection = new TCPMasterConnection(InetAddress.getByName(uri.host()));
			connection.setPort(Integer.parseInt(uri.port() == null ? "502" : uri.port()));	
			connection.connect();
		}
		
		return connection;
	}
	
	private synchronized InputRegister[] readMultipleEntries(int register, int length) throws Exception {

		ReadInputRegistersRequest req = new ReadInputRegistersRequest(register, length);
		req.setUnitID(unitID);
		
		ModbusTCPTransaction trans = this.getModbusTCPTransaction();
		trans.setRequest(req);
		trans.execute();
		ReadInputRegistersResponse res = (ReadInputRegistersResponse) trans.getResponse();
		
		// Synchronisierung mit Transaction
		if(res.getTransactionID() != trans.getTransactionID()) {
			connection.close();
			throw new Exception("Transaction IDs not identical: TransactionID=" + trans.getTransactionID() + ", ResponseID=" + res.getTransactionID());
		}
		
		return res.getRegisters();
	}
	
	private synchronized BitVector readMultipleEntriesDiscrete(int register, int bitSize) throws Exception {

		ReadInputDiscretesRequest requestConstant = new ReadInputDiscretesRequest(register, bitSize);
		requestConstant.setUnitID(unitID);
		ModbusTCPTransaction trans = this.getModbusTCPTransaction();
		trans.setRequest(requestConstant);
		
		try {
			trans.execute();
		} catch(ModbusIOException e) {
			connection.close();
			throw new ModbusIOException("Connection to device lost");
		}
		
		ReadInputDiscretesResponse res = (ReadInputDiscretesResponse) trans.getResponse();
		
		// Synchronisierung mit Transaction
		if(res.getTransactionID() != trans.getTransactionID()) {
			connection.close();
			throw new Exception("Transaction IDs not identical: TransactionID=" + trans.getTransactionID() + ", ResponseID=" + res.getTransactionID());
		}
		
		// Byte als unsigned interpretieren
		return res.getDiscretes();
	}
	
	private synchronized void writeMultipleEntries(int address, Register[] registers) throws Exception {
		
		WriteMultipleRegistersRequest req = new WriteMultipleRegistersRequest(address, registers);
		req.setUnitID(unitID);
		
		ModbusTCPTransaction trans = this.getModbusTCPTransaction();
		trans.setRequest(req);
		try {
			trans.execute(); 
			WriteMultipleRegistersResponse res = (WriteMultipleRegistersResponse) trans.getResponse();
			if(res.getTransactionID() != trans.getTransactionID()) {
				connection.close();
				throw new Exception("Transaction IDs not identical: TransactionID=" + trans.getTransactionID() + ", ResponseID=" + res.getTransactionID());
			}
			
		} catch (ModbusIOException e) {
			connection.close();
			throw new ModbusIOException("Connection to device lost");
		}
	}
	
	private synchronized void writeMultipleEntriesDiscrete(int register, BitVector bitVector) throws Exception {
		
		WriteMultipleCoilsRequest req = new WriteMultipleCoilsRequest(register, bitVector);
		req.setUnitID(unitID);
		
		ModbusTCPTransaction trans = this.getModbusTCPTransaction();
		trans.setRequest(req);
		try {
			trans.execute();
			WriteMultipleCoilsResponse res = (WriteMultipleCoilsResponse) trans.getResponse();
			if(res.getTransactionID() != trans.getTransactionID()) {
				connection.close();
				throw new Exception("Transaction IDs not identical: TransactionID=" + trans.getTransactionID() + ", ResponseID=" + res.getTransactionID());
			}
			
		}catch (ModbusIOException e) {
			throw new ModbusIOException("Connection to device lost");
		}
	}

	@Override
	public void parseAddressString(RealProperty realProperty) throws Exception {
		AddressParser addressParser = new AddressParser();
		if(uri.segmentCount() >= 1) {
			if("wago".equals(uri.segment(0)))
				addressParser.setModbusType(AddressParser.MODBUS_TYPE_WAGO);
			else if("default".equals(uri.segment(0)))
				addressParser.setModbusType(AddressParser.MODBUS_TYPE_DEFAULT);
		}
		
		addressParser.parseAddress(realProperty.getDriverAddress());
		realProperty.setUserDefinedAddressObject(addressParser);
	}

	@Override
	public int readData(DataTable dataTable) throws Exception {
		RealProperty firstProp = dataTable.getProperties().get(0);
		AddressParser firstPropAP = (AddressParser) firstProp.getUserDefinedAddressObject();
		
		if(firstPropAP.isDiscrete()) {
			ByteBuffer bb = dataTable.getByteBuffer();
			BitVector bv = readMultipleEntriesDiscrete(dataTable.getAbsoluteAddress(), dataTable.getProperties().size());
			for(RealProperty property : dataTable.getProperties()) {
				AddressParser addressParser = (AddressParser) property.getUserDefinedAddressObject();
				if(!addressParser.isDiscrete())
					throw new Exception(property.getName() + " has to be discrete: Illegal definition in SignalList");
				bb.position(property.getAddressOffset());
				bb.putShort((short) (bv.getBit(addressParser.getReadAddress()) ? 1 : 0));
			}
		}
		else {
			InputRegister[] inputRegisters = this.readMultipleEntries(firstPropAP.getReadAddress(), firstPropAP.getLength());
			ByteBuffer bb = dataTable.getByteBuffer();
			bb.position(firstProp.getAddressOffset());
			
			for(int i=0; i<inputRegisters.length; ++i) {
				if(firstProp instanceof PropertyInt32) {
					// Wrter mssen getauscht werden
					byte[] wordLow = inputRegisters[i].toBytes();
					byte[] wordHigh = inputRegisters[i+1].toBytes();
					bb.put(wordHigh);
					bb.put(wordLow);
					++i;
				}
				else {
					bb.put(inputRegisters[i].toBytes());
				}
			}	
		}			
		return 0;
	}
}

