package ch.thomson.ucs.tine.server.internal.client;

import java.net.InetAddress;
import java.util.HashSet;
import java.util.Set;

import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.EquipmentModule;
import ch.thomson.ucs.tine.server.FrontEndController;
import ch.thomson.ucs.tine.server.Observable;
import ch.thomson.ucs.tine.server.Property;
import ch.thomson.ucs.tine.server.PropertyNumeric;
import ch.thomson.ucs.tine.server.util.ServerResourceLoader;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;

public class PropertyNumericDAO extends PropertyValueProvider {

	protected float[] readBufferFloat;
	protected float[] writeBufferFloat;
	protected TDataType doutFloat;
	protected TDataType dinFloat;
	
	protected int[] readBufferInt;
	protected int[] writeBufferInt;
	protected TDataType doutInt;
	protected TDataType dinInt;
	
	protected long lastTimeStamp;
	
	protected Set<PropertyNumericChangeListener> listeners = new HashSet<PropertyNumericChangeListener>();
	
	protected PropertyNumericDAO(Device device, PropertyNumeric property, boolean notificated) throws Exception {
		super(device, property, notificated);
		
		readBufferFloat = new float[property.getLength()];
		doutFloat = new TDataType(readBufferFloat);
		writeBufferFloat = new float[property.getLength()];
		dinFloat = new TDataType(writeBufferFloat);
		
		readBufferInt = new int[property.getLength()];
		doutInt = new TDataType(readBufferInt);
		writeBufferInt = new int[property.getLength()];
		dinInt = new TDataType(writeBufferInt);
		
		refreshValues();
	}
	
	@Override
	public long getTimeStamp() {
		return lastTimeStamp;
	}
	
	protected void refreshValuesFloat() {
		lastTineErrorCode = property.readData(device, doutFloat);
		lastTimeStamp = doutFloat.getDataTimeStamp();
	}
	
	protected void refreshValuesInt() {
		lastTineErrorCode = property.readData(device, doutInt);
		lastTimeStamp = doutInt.getDataTimeStamp();
	}
	
	protected void writeValuesFloat(int length) {
		dinFloat.putData(writeBufferFloat);
		dinFloat.dArrayLength = length;
		lastTineErrorCode = property.writeData(device, dinFloat);
	}
	
	protected void writeValuesInt(int length) {
		dinInt.putData(writeBufferInt);
		dinInt.dArrayLength = length;
		lastTineErrorCode = property.writeData(device, dinInt);
	}
	
	protected void writeValuesFloat() {
		writeValuesFloat(getLength());
	}
	
	protected void writeValuesInt() {
		writeValuesInt(getLength());
	}
	
	@Override
	protected void refreshValues() {
		refreshValuesFloat();
		refreshValuesInt();
	}

	@Override
	protected void writeValues(int length) {
		throw new UnsupportedOperationException();
	}

	@Override
	protected void writeValues() {
		throw new UnsupportedOperationException();
	}

	private static InetAddress inetAddress;
	
	/**
	 * 
	 * @param tinePath Syntax: Absoluter Pfad "/CONTEXT/EQM/DEVICE/PROPERTY" oder relativer Pfad "EQM/DEVICE/PROPERTY"
	 * @param notificated Bei true wird das DAO informiert, wenn sich der Wert der Property gendert hat.
	 * @return Datenzugriffspunkt
	 * @throws Exception Bei falschem Pfad
	 */
	public static PropertyNumericDAO createPropertyNumericDAO(String tinePath, boolean notificated) throws Exception {
		if(!tinePath.startsWith("/")) {
			if(inetAddress == null) {
				inetAddress = InetAddress.getLocalHost();
			}
			
			tinePath = "/" + inetAddress.getHostName() + "/" + tinePath;
		}
		
		String[] segments = tinePath.split("/");
		if(segments.length != 5)
			throw new Exception("Illegal Property Path: " + tinePath);
		
		// Es gibt nur einen FEC...
		FrontEndController fec = ServerResourceLoader.getFrontEndControllerInstance();
		fec.initialize();
				
		EquipmentModule eqm = fec.findEquipmentModule(segments[2]);
		if(eqm == null)
			throw new Exception("Equipment Module " + segments[2] + " not found");
		
		Device device = eqm.findDevice(segments[3]);
		if(device == null) 
			throw new Exception("Device " + segments[3] + " not found");
		
		
		Property p = eqm.findProperty(segments[4]);
		if(p == null)
			throw new Exception("Property " + segments[4] + " not found");
		
		if(!(p instanceof PropertyNumeric)) 
			throw new Exception("Property " + segments[4] + " is not numeric");
		
		return new PropertyNumericDAO(device, (PropertyNumeric)p, notificated);
	}
	
	public static PropertyNumericDAO createPropertyNumericDAO(Device device, Property p, boolean notificated) throws Exception {
		if(!(p instanceof PropertyNumeric))
			throw new Exception("Property is not numeric");
		
		return new PropertyNumericDAO(device, (PropertyNumeric)p, notificated);
	}
	
	@Override
	public void update(Observable observable) {
		if(observable == property) {
			float[] oldValuesFloat = readBufferFloat.clone();
			int[] oldValuesInt = readBufferInt.clone();
			refreshValues();
			
			PropertyNumericNotification n = new PropertyNumericNotification(this, oldValuesFloat, getValuesAsFloat(), oldValuesInt, getValuesAsInt());
			for(PropertyNumericChangeListener l : listeners)
				l.notifyChanged(n);
		}
	}
	
	public void addChangeListener(PropertyNumericChangeListener l) {
		this.listeners.add(l);
	}
	
	public void removeChangeListener(PropertyNumericChangeListener l) {
		this.listeners.remove(l);
	}
	
	public int getInt() {
		return getValuesAsInt()[0];
	}
	
	public void setInt(int value) {
		writeBufferInt[0] = value;
		writeValuesInt(1);
	}
	
	public float getFloat() {
		return getValuesAsFloat()[0];
	}
	
	public void setFloat(float value) {
		writeBufferFloat[0] = value;
		writeValuesFloat(1);
	}
	
	public void setValuesAsInt(int[] values) {
		System.arraycopy(values, 0, writeBufferInt, 0, values.length);
		writeValuesInt(values.length);
	}
	
	public int[] getValuesAsInt() {
		if(isNotificated())
			return readBufferInt; // gecachter Wert zurckgeben
		
		refreshValuesInt();
		return readBufferInt;
	}
	
	public void setValuesAsFloat(float[] values) {
		System.arraycopy(values, 0, writeBufferFloat, 0, values.length);
		writeValuesFloat(values.length);
		if(lastTineErrorCode != TErrorList.success)
			System.err.println(getProperty().getName() + " setValuesAsFloat failed: " + TErrorList.getErrorString(lastTineErrorCode));
	}
	
	public float[] getValuesAsFloat() {
		
		if(isNotificated())
			return readBufferFloat; // gecachter Wert zurckgeben
		
		refreshValuesFloat();
		return readBufferFloat;
	}
	
	public void setInt(int index, int value) {
		// TODO: Diese Methode ist potenziell gefhrlich. Sie liest das Array,
		// manipuliert es, und schreibt es zurck. Es ist aber nicht
		// sichergestellt, ob die Werte in der Zwischenzeit gendert haben.
		System.arraycopy(getValuesAsInt(), 0, writeBufferInt, 0, getLength());
		writeBufferInt[index] = value;
		writeValuesInt();
	}
	
	public int getInt(int index) {
		return getValuesAsInt()[index];
	}
	
	public float getFloat(int index) {
		return getValuesAsFloat()[index];
	}	
}
