package ch.thomson.ucs.tine.server.internal.client;

import java.net.InetAddress;

import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.EquipmentModule;
import ch.thomson.ucs.tine.server.FrontEndController;
import ch.thomson.ucs.tine.server.Property;
import ch.thomson.ucs.tine.server.PropertyNumeric;
import ch.thomson.ucs.tine.server.util.ServerResourceLoader;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;

public abstract class PropertyValueProvider extends ObserverAdapter {
	Device device;
	PropertyNumeric property;
	boolean notificated;
	TDataType dout;
	TDataType din;
	int lastTineErrorCode = TErrorList.connection_timeout;
	
	public PropertyValueProvider(Device device, PropertyNumeric property, boolean notificated) {
		this.device = device;
		this.property = property;
		this.notificated = notificated;
		
		if(notificated)
			property.addObserver(this);
	}
	
	protected void refreshValues() {
		lastTineErrorCode = property.readData(device, dout);
	}
	
	protected void writeValues(int length) {
		din.dArrayLength = length;
		lastTineErrorCode = property.writeData(device, din);
	}
	
	protected void writeValues() {
		writeValues(getLength());
	}
	
	public boolean isNotificated() {
		return notificated;
	}
	
	public void dispose() {
		property.removeObserver(this);
	}
	
	public void startTransaction() {
		this.device.startTransaction();
	}
	
	public void commit() throws Exception {
		this.device.commit();
	}
	
	public int getLength() {
		return property.getLength();
	}

	public Device getDevice() {
		return device;
	}

	public PropertyNumeric getProperty() {
		return property;
	}
	
	public String getEquipmentName() {
		return property.getEquipmentModule().getExportName();
	}
	
	public void setScaleFactor(double value) {
		property.setFactor(value);
		this.update(property);
	}

	public long getTimeStamp() {
		return dout.getDataTimeStamp();
	}
	
	public String getErrorString() {
		return TErrorList.getErrorString(lastTineErrorCode);
	}
	
	public boolean hasError() {
		return lastTineErrorCode != TErrorList.success;
	}
	
	private static class TineAddress {
		public Device device;
		public PropertyNumeric property;
	}
	
	protected static InetAddress inetAddress;
	protected static TineAddress getTineAddress(String tinePath) throws Exception {
		if(!tinePath.startsWith("/")) {
			if(inetAddress == null) {
				inetAddress = InetAddress.getLocalHost();
			}
			
			tinePath = "/" + inetAddress.getHostName() + "/" + tinePath;
		}
		
		String[] segments = tinePath.split("/");
		if(segments.length != 5)
			throw new Exception("Illegal Property Path: " + tinePath);
		
		// Es gibt nur einen FEC...
		FrontEndController fec = ServerResourceLoader.getFrontEndControllerInstance();
		fec.initialize();
				
		EquipmentModule eqm = fec.findEquipmentModule(segments[2]);
		if(eqm == null)
			throw new Exception("Equipment Module " + segments[2] + " not found");
		
		Device device = eqm.findDevice(segments[3]);
		if(device == null) 
			throw new Exception("Device " + segments[3] + " not found");
		
		
		Property p = eqm.findProperty(segments[4]);
		if(p == null)
			throw new Exception("Property " + segments[4] + " not found");
		
		if(!(p instanceof PropertyNumeric)) 
			throw new Exception("Property " + segments[4] + " is not numeric");
		
		TineAddress ta = new TineAddress();
		ta.device = device;
		ta.property = (PropertyNumeric) p;
		
		return ta;
	}
	
	/**
	 * 
	 * @param tinePath
	 *            Syntax: Absoluter Pfad "/CONTEXT/EQM/DEVICE/PROPERTY" oder
	 *            relativer Pfad "EQM/DEVICE/PROPERTY"
	 * @param notificated
	 *            Bei true wird das DAO informiert, wenn sich der Wert der
	 *            Property gendert hat.
	 * @return Datenzugriffspunkt
	 * @throws Exception
	 *             Bei falschem Pfad
	 */
	public static PropertyIntValueProvider createPropertyIntValueProvider(String tinePath, boolean notificated) throws Exception {
		TineAddress ta = getTineAddress(tinePath);
		return createPropertyIntValueProvider(ta.device, ta.property, notificated);
	}
	
	public static PropertyIntValueProvider createPropertyIntValueProvider(Device device, Property p, boolean notificated) throws Exception {
		if(!(p instanceof PropertyNumeric))
			throw new Exception("Property " + p.getName()+ "is not numeric");
		return new PropertyIntValueProvider(device, (PropertyNumeric) p, notificated);
	}
}
