package ch.thomson.ucs.tine.server.util;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.ini4j.Ini;
import org.ini4j.InvalidFileFormatException;

import ch.thomson.ucs.tine.server.GroupProperty;
import ch.thomson.ucs.tine.server.Property;

public class UCSConfig {
	private static final String sectionNameTine = "TINE";
	private static final String iniPathEnv = "TINE_HOME";
	private static final String iniName = "ucs.ini";
	
	private Map<String, String> eqmNameExternalToInternalMap;
	private Map<String, String> eqmNameInternalToExternalMap;
	private Map<String, String> deviceNameExternalToInternalMap;
	private Map<String, String> deviceNameInternalToExternalMap;
	private Map<String, String> propNameExternalToInternalMap;
	private Map<String, String> propNameInternalToExternalMap;
	
	
	private static UCSConfig instance = null;
	private Ini ini = null;
	private Boolean iniFileExisting = null;
	
	private Logger logger = Logger.getLogger("ucsLogger");
	
	public String usedFecContextName;
	
	private UCSConfig() {}
	
	public boolean hasIniFile() {
		return getIni() != null;
	}
	
	public static synchronized UCSConfig getInstance() {
		if(instance == null) {
			instance = new UCSConfig();
		}
		return instance;
	}
	
	private Ini getIni() {
		if (iniFileExisting == null) {
			String path = System.getenv(iniPathEnv);
			File file = null;
			if (path != null) {
				try {
					file = new File(path + File.separator + iniName);
					ini = new Ini(file);
					iniFileExisting = true;
					logger.info( file.getAbsolutePath() + " found, will work with this file's TINE Namespace Settings.");
				} catch (InvalidFileFormatException e) {
					iniFileExisting = false;
					logger.error(file.getAbsolutePath() + " not found", e);
				} catch (IOException e) {
					iniFileExisting = false;
					logger.info(file.getAbsolutePath() + " not found, will work with INTERNAL TINE Namespace Settings.");
				}
			}
		}
		
		return ini;
	}
	
	private String askIniFile(String section, String key) {
		if(getIni() == null)
			return null; // Es existiert kein Ini-File
		
		return ini.fetch(section, key);
	}
	
	public String getContextName() {	
		return askIniFile(sectionNameTine, "CONTEXT");
	}

	public String getFecName() {
		return askIniFile(sectionNameTine, "FEC_NAME");
	}
	
	public String getFecSubSystem() {
		return askIniFile(sectionNameTine, "SUBSYSTEM");
	}
	
	public String getFecDescription() {
		String s = askIniFile(sectionNameTine, "DESCRIPTION");
		return s != null ? s : "unknown";
	}
	
	public String getFecResponsible() {
		String s = askIniFile(sectionNameTine, "RESPONSIBLE");
		return s != null ? s : "unknown";
	}
	
	public String getFecLocation() {
		String s = askIniFile(sectionNameTine, "LOCATION");
		return s != null ? s : "unknown";
	}

	/**
	 * Gibt, falls ein ini-File vorhanden ist und ein anderer Device-Name dort
	 * definiert wurde, diesen zurck, ansonsten wird der bergebene Device-Name
	 * wieder zurckgeben. <br/>Der Name wird IMMER, auch wenn es nichts zu ersetzen
	 * gibt, in den entsprechenden HashMaps eingetragen, damit die bersetzung
	 * jederzeit zur Verfgung steht.
	 * 
	 * @param device
	 *            Device-Name, der gegebenfalls ersetzt werden muss.
	 * @return Device-Name
	 */
	public String getDeviceName(String device) {
		String s = askIniFile(sectionNameTine, device.toUpperCase());
		s = s != null ? s : device;
		getDeviceNameExternalToInternalMap().put(s, device);
		getDeviceNameInternalToExternalMap().put(device, s);
		return s;
	}
	
	/**
	 * Gibt, falls ein ini-File vorhanden ist und ein anderer EQM-Name dort
	 * definiert wurde, diesen zurck, ansonsten wird der bergebene EQM-Name
	 * wieder zurckgeben. <br/>
	 * Der Name wird IMMER, auch wenn es nichts zu ersetzen gibt, in den
	 * entsprechenden HashMaps eingetragen, damit die bersetzung jederzeit zur
	 * Verfgung steht.
	 * 
	 * @param eqm
	 *            EQM-Name, der gegebenfalls ersetzt werden muss.
	 * @return EQM-Name
	 */
	public String getEQMExportName(String eqm) {
		String s = askIniFile(sectionNameTine, "EQM_EXPORT_" + eqm);
		s = s != null ? s : eqm;
		getEqmNameExternalToInternalMap().put(s, eqm);
		getEqmNameInternalToExternalMap().put(eqm, s);
		return s;
	}
	
	public String getEQMLocalName(String eqm) {
		String s = askIniFile(sectionNameTine, "EQM_LOCAL_" + eqm);
		return s != null ? s : eqm;
	}
	
	/**
	 * Gibt, falls ein ini-File vorhanden ist und ein Property Prefix dort
	 * definiert wurde, den bergebenen Namen inkl. Prefix zurck, ansonsten
	 * wird der bergebene Property-Name wieder zurckgeben. <br/>
	 * Der Name wird IMMER, auch wenn es nichts zu ersetzen gibt, in den
	 * entsprechenden HashMaps eingetragen, damit die bersetzung jederzeit zur
	 * Verfgung steht.
	 * 
	 * @param propName
	 *            Property-Name, der gegebenfalls ersetzt werden muss.
	 * @return Property-Name
	 */
	public String getPropertyNamePrefixed(Property p) {
		String s = askIniFile(sectionNameTine, "PROPERTY_INTERNAL_PREFIX");
		s = (s != null && isPropertyInternal(p)) ? (s + p.getName()) : p.getName();
		getPropNameExternalToInternalMap().put(s, p.getName());
		getPropNameInternalToExternalMap().put(p.getName(), s);
		return s;
	}
	
	public boolean isPropertyInternal(Property p) {
		return (p instanceof GroupProperty || p.getLogonLevel() > 1);
	}

	public Map<String, String> getEqmNameExternalToInternalMap() {
		if (eqmNameExternalToInternalMap == null)
			eqmNameExternalToInternalMap = new HashMap<String, String>();
		return eqmNameExternalToInternalMap;
	}
	
	public Map<String, String> getEqmNameInternalToExternalMap() {
		if (eqmNameInternalToExternalMap == null)
			eqmNameInternalToExternalMap = new HashMap<String, String>();
		return eqmNameInternalToExternalMap;
	}

	public Map<String, String> getDeviceNameExternalToInternalMap() {
		if(deviceNameExternalToInternalMap == null)
			deviceNameExternalToInternalMap = new HashMap<String, String>();
		return deviceNameExternalToInternalMap;
	}
	
	public Map<String, String> getDeviceNameInternalToExternalMap() {
		if(deviceNameInternalToExternalMap == null)
			deviceNameInternalToExternalMap = new HashMap<String, String>();
		return deviceNameInternalToExternalMap;
	}
	
	public Map<String, String> getPropNameExternalToInternalMap() {
		if(propNameExternalToInternalMap == null)
			propNameExternalToInternalMap = new HashMap<String, String>();
		return propNameExternalToInternalMap;
	}
	
	public Map<String, String> getPropNameInternalToExternalMap() {
		if(propNameInternalToExternalMap == null)
			propNameInternalToExternalMap = new HashMap<String, String>();
		return propNameInternalToExternalMap;
	}
}
