package ch.thomson.ucs.utils.logging;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;
import org.osgi.framework.Bundle;

/**
 * Diese Klasse erleichtert die Benutzung der Eclipse Logging API. Die Instanz wird ber eine der beiden Factory-Methoden erzeugt. Beispiel:<br>
 * Bundle:<br> <code>IPluginLogger logger = PluginLogger.getLogger(Activator.getContext().getBundle());</code><br><br>
 * Plugin:<br> <code>IPluginLogger logger = PluginLogger.getLogger(Activator.getDefault());</code>
 * @author Urs Fischer
 *
 */
public final class PluginLogger implements IPluginLogger {

	private Bundle bundle;
	private boolean debugEnabled;

	public static IPluginLogger getLogger(Plugin runtimePlugin) {
		return new PluginLogger(runtimePlugin.getBundle(), runtimePlugin.isDebugging());
	}
	
	public static IPluginLogger getLogger(Bundle bundle) {
		return new PluginLogger(bundle, Platform.inDebugMode());
	}

	private PluginLogger(Bundle bundle, boolean debugEnabled) {
		super();
		this.bundle = bundle;
		this.debugEnabled = debugEnabled;
	}

	@Override
	public final void info(String message, Throwable throwable) {
		log(IStatus.INFO, IStatus.OK, message, throwable);
	}

	@Override
	public final void info(String message) {
		log(IStatus.INFO, IStatus.OK, message, null);
	}

	@Override
	public final void error(String message) {
		log(IStatus.ERROR, IStatus.OK, message, null);
	}

	@Override
	public final void error(Throwable throwable) {
		error(throwable.getMessage(), throwable);
	}

	@Override
	public final void error(String message, Throwable throwable) {
		log(IStatus.ERROR, IStatus.OK, message, throwable);
	}

	private final void log(int severity, int code, String message,
			Throwable exception) {
		if (severity == IStatus.ERROR || (debugEnabled)) {
			log(createStatus(severity, code, message, exception));
		}
	}

	private final IStatus createStatus(int severity, int code, String message,
			Throwable exception) {
		String symbolicName = bundle.getSymbolicName();
		return new Status(severity, symbolicName, code,
				message != null ? message : "", exception); //$NON-NLS-1$
	}

	private final void log(IStatus status) {
		Platform.getLog(bundle).log(status);
	}
}
