package ch.thomson.ucs.utils.logic;

import java.io.Closeable;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.Flushable;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteOrder;

/**
 * Diese Klasse dekoriert die Java-Klasse DataOutputStream, da eine direkte
 * Vererbung aufgrund der als Final deklarierten Methoden nicht mglich ist.
 * Standardmssig interpretiert die Java-Klasse die Byte Order als Big Endian.
 * Mit Hilfe dieser Erweiterung kann dies beliebig umgestellt werden.
 * 
 * @author FischerU
 *
 */
public class ByteOrderedDataOutputStream implements DataOutput, Flushable, Closeable {
	public static final int BIG_ENDIAN = 0;
	public static final int LITTLE_ENDIAN = 1;
	
	protected ByteOrder byteOrder = ByteOrder.BIG_ENDIAN; // Default
	protected DataOutputStream dos;
	
	public ByteOrderedDataOutputStream(OutputStream os) {
		dos = new DataOutputStream(os);
	}
	
	public ByteOrderedDataOutputStream(OutputStream os, ByteOrder byteOrder) {
		this(os);
		this.byteOrder = byteOrder;
	}

	public ByteOrder getByteOrder() {
		return byteOrder;
	}

	public void setByteOrder(ByteOrder byteOrder) {
		this.byteOrder = byteOrder;
	}

	@Override
	public void write(int b) throws IOException {
		dos.write(b);
	}

	@Override
	public void write(byte[] b) throws IOException {
		dos.write(b);
	}

	@Override
	public void write(byte[] b, int off, int len) throws IOException {
		dos.write(b, off, len);
	}

	@Override
	public void writeBoolean(boolean v) throws IOException {
		dos.writeBoolean(v);
	}

	@Override
	public void writeByte(int v) throws IOException {
		dos.writeByte(v);
	}

	@Override
	public void writeBytes(String s) throws IOException {
		dos.writeBytes(s);
	}

	@Override
	public void writeChar(int v) throws IOException {
		this.writeShort(v);
	}

	@Override
	public void writeChars(String s) throws IOException {
		dos.writeChars(s);
	}

	@Override
	public void writeDouble(double v) throws IOException {
		dos.writeDouble(byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(v) : v);
	}

	@Override
	public void writeFloat(float v) throws IOException {
		dos.writeFloat(byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(v) : v);
	}

	@Override
	public void writeInt(int v) throws IOException {
		dos.writeInt(byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(v) : v);
	}

	@Override
	public void writeLong(long v) throws IOException {
		dos.writeLong(byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(v) : v);
	}

	@Override
	public void writeShort(int v) throws IOException {
		dos.writeShort(byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap((short)v) : v);
	}

	@Override
	public void writeUTF(String s) throws IOException {
		dos.writeUTF(s);
	}

	@Override
	public void close() throws IOException {
		dos.close();
	}

	@Override
	public void flush() throws IOException {
		dos.flush();
	}
}
