/**
 * 
 */
package de.desy.acop.video.analysis;


import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

import de.desy.acop.displayers.tools.DismissableBlockingQueue;
import de.desy.acop.displayers.tools.TINEThreadFactory;
import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.video.VideoServerConnection;
import de.desy.acop.video.analysis.gaussian.AImage;
import de.desy.acop.video.analysis.gaussian.GaussianImageAnalyzer;
import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.displayer.ImageFormat;
import de.desy.acop.video.displayer.TineHandler;
import de.desy.acop.video.displayer.TineImageReceiver;
import de.desy.acop.video.displayer.timage.TImageConverter;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;

/**
 * </code>ImageAnalysisEngine</code> provides </code>AImage</code> image analysis 
 * data. It has two modes of operation: local and remote analysis. In local analysis
 * mode, the </code>IMAGE</code> arguments of the </code>updateValue</code> method
 * are processed locally. In remote mode this engine reads </code>AImage</code> data
 * from a video analysis server. 
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
@SuppressWarnings("deprecation")
public class ImageAnalysisEngine implements TineImageReceiver {
	
	public static final String PROPERTY_A_IMAGE = "aImage";
	public static final String PROPERTY_BACKGROUND = "background";
	public static final String PROPERTY_PRECISE_BACKGROUND = "preciseBackground";
	public static final String PROPERTY_RESET = "reset";
		
	private PropertyChangeSupport pcSupport;
	private Executor executor;
	private ThreadPoolExecutor analysisExecutor;
	private boolean localAnalysis = true;
	
	private VideoServerConnection videoServerConnection;
	
	protected ColorDecoder decoder;
	
	private AnalysisImage aImage;
	private IMAGE image;
	private IMAGE background;
	private double[] preciseBackground;
	private int imageWidth = -1;
	private int imageHeight = -1;
	private int backgroundWidth = -1;
	private int backgroundHeight = -1;
	private int roiX = -1;
	private int roiY = -1;
	private int roiW = -1;
	private int roiH = -1;
	private int roi2X = -1;
	private int roi2Y = -1;
	private int roi2W = -1;
	private int roi2H = -1;
	private double threshold = Double.NaN;
	private boolean calculateThreshold = true;
	private boolean performFit = false;
	private boolean performSmoothing = false;
	private ColorMap colorMap = null;
	
	private ImageAnalyzer imageAnalyzer = new GaussianImageAnalyzer();
	/**
	 * Constructs this </code>ImageAnalysisEngine</code>. 
	 * @param decoder the </code>ColorDecoder</code> to be used with this 
	 * </code>ImageAnalysisEngine</code>
	 */
	public ImageAnalysisEngine(ColorDecoder decoder) {
		this.decoder = decoder;
		pcSupport = new PropertyChangeSupport(this);
	}
		
	@Override
	public void resetForReceiving() {
		pcSupport.firePropertyChange(PROPERTY_RESET, false, true);
	}

	@Override
	public synchronized void updateValue(IMAGE newImage) {
		if (!localAnalysis) return;
//		logger.log(Level.INFO,"New update received.");
//		image = TImageUtils.clone(newImage); // mdavid: commented
		image = newImage.clone();

		FrameHeader hdr = image.getFrameHeader();
		int w = hdr.aoiWidth == -1 ? hdr.sourceWidth : hdr.aoiWidth;
		int h = hdr.aoiHeight == -1 ? hdr.sourceHeight : hdr.aoiHeight;
		
		if (w != imageWidth || h != imageHeight) {
			roiX = 0;
			roiY = 0;
			roiW = w;
			roiH = h;
			roi2X = 0;
			roi2Y = 0;
			roi2W = w;
			roi2H = h;
		}
		
		imageWidth = w;
		imageHeight = h;

		if ((imageWidth != backgroundWidth || imageHeight != backgroundHeight) && background != null) {
			setBackground(null);
		}
		else performAnalysis();
	}
	
	/**
	 * Sets the precise background array. This is the array that holds values per pixels
	 * for whole background image, but instead of int values, pixel values are given as
	 * luminosity doubles. Negative values are permitted and will be taken into account
	 * during analysis, but forced to 0, when real image is constructed.
	 * 
	 * @param background the pixel value array
	 * @param width the width of the image
	 */
	public synchronized void setPreciseBackground(double[] background, int width) {
		TineHandler.logger.log(Level.INFO, "Background image set.");
		setPreciseBackground(background,width, true);
	}
	
	/**
	 * Returns the double array that is used as background image.
	 * 
	 * @see #setPreciseBackground(double[], boolean)
	 * @return the double array image
	 */
	public synchronized double[] getPreciseBackground() {
		return preciseBackground;
	}
	
	/**
	 * Returns the width of the background image.
	 * 
	 * @return the width
	 */
	public synchronized int getBackgroundWidth() {
		return backgroundWidth;
	}
	
	/**
	 * Sets the background image and initiates required actions.
	 * 
	 * @param background the background image array
	 * @param width the width of the image
	 * @param reanalyze true if the image should be reanalyzed
	 */
	private synchronized void setPreciseBackground(final double[] preciseBackground, final int width, boolean reanalyze) {
		this.background = null;
		this.preciseBackground = null;
		backgroundWidth = -1;
		backgroundHeight = -1;
		if (preciseBackground != null) {
			int w = width;
			int h = preciseBackground.length/width;
			
			if (w == imageWidth && h == imageHeight) {
				backgroundWidth = w;
				backgroundHeight = h;
				this.preciseBackground = preciseBackground;
				this.background = toImage(preciseBackground,width);
			}
		}
		
		if (reanalyze) {
			if (localAnalysis) performAnalysis();
			else {
				executeTask(new Runnable() {
					@Override
					public void run() {
						try {
							getVideoServerConnection().setBackground(background);
						} catch (ConnectionFailed e) {
							e.printStackTrace();
						}
					}
				});
			}
			pcSupport.firePropertyChange(PROPERTY_PRECISE_BACKGROUND, null, this.preciseBackground);
			pcSupport.firePropertyChange(PROPERTY_BACKGROUND, null, this.background);
		}
	}
	
	/**
	 * Transforms the double array to a Gray image.
	 * 
	 * @param image the image to transform 
	 * @param width the width of the image
	 * @return generated IMAGE
	 */
	public static IMAGE toImage(double[] image,int width) {
		if (image == null || image.length == 0) return null;
		if (width < 1 || image.length/width < 1) return null;
		BufferedImage bi = toBufferedImage(image,width);
//		return TBufferedImage.toIMAGE(bi);
		return TImageConverter.parseBufferedImage(bi);
	}	
	
	/**
	 * Transforms teh double array to a gray buffered image.
	 * 
	 * @param image the double image array 
	 * @param width the width of the image
	 * @return buffered image
	 */
	public static BufferedImage toBufferedImage(double[] image,int width) {
		if (image == null || image.length == 0) return null;
		if (width < 1 || image.length/width < 1) return null;
		int i = 0;
		int length = image.length;
		int[] rgb = new int[length];
		try {
			int alpha = (255 << 24);
			int val;
    		for (;;) {
    			val = (int)(image[i]+0.5);
    			if (val < 0) val = 0;
    			rgb[i] =  alpha + (val << 16) + (val << 8) + (val);
    			i++;
    		}
		} catch (ArrayIndexOutOfBoundsException e) {
			//ignore
		}
		BufferedImage bi = new BufferedImage(width,length/width,BufferedImage.TYPE_BYTE_GRAY);
		
		bi.setRGB(0,0,width,length/width,rgb,0,width);
		return bi;
	}	
	
	/**
	 * Gets the background image used by this </code>ImageAnalysisEngine</code>.
	 * @return the background image
	 */
	public synchronized IMAGE getBackground() {
		return background;
	}
		
	/**
	 * Sets the background image to be used by this </code>ImageAnalysisEngine</code>.
	 * @param background the image to set
	 */
	public synchronized void setBackground(IMAGE background) {
		TineHandler.logger.log(Level.INFO, "Background image set.");
		setBackground(background, true);
		this.background = background;
	}
	
	private synchronized void setBackground(final IMAGE background, boolean reanalyze) {
		if (background == null) {
			setPreciseBackground(null,-1);
		} else {
			int[] rgb = toBufferedImage(background,colorMap).getRGB(0, 0, background.getFrameHeader().sourceWidth, 
					background.getFrameHeader().sourceHeight, null, 0, background.getFrameHeader().sourceWidth);
			double[] data = getDecoder().transform(rgb);
			setPreciseBackground(data,background.getFrameHeader().sourceWidth,reanalyze);
			this.background = background;
		}
	}
	
	/**
	 * Sets the region of interest and reanalyzes pixels.
	 * @param roiX the starting x coordinate
	 * @param roiY the starting y coordinate
	 * @param roiW the width of the region
	 * @param roiH the height of the region
	 */
	public synchronized void setRoi(final int roiX, final int roiY, final int roiW, final int roiH) {
		if (this.roiX == roiX && this.roiY == roiY && this.roiW == roiW && this.roiH == roiH) return;
		TineHandler.logger.log(Level.INFO, "ROI: x="+roiX+", y="+roiY +", w="+roiW + ", h="+roiH);
		setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Sets the region of interest and reanalyzes pixels.
	 * @param roiX the starting x coordinate
	 * @param roiY the starting y coordinate
	 * @param roiW the width of the region
	 * @param roiH the height of the region
	 */
	public synchronized void setRoi2(final int roi2X, final int roi2Y, final int roi2W, final int roi2H) {
		if (this.roi2X == roi2X && this.roi2Y == roi2Y && this.roi2W == roi2W && this.roi2H == roi2H) return;
		TineHandler.logger.log(Level.INFO, "Threshold ROI: x="+roi2X+", y="+roi2Y +", w="+roi2W + ", h="+roi2H);
		setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Gets the </code>IMAGE</code>.
	 * @return the </code>IMAGE</code>
	 */
	public synchronized IMAGE getImage() {
		return image;
	}
	
	/**
	 * Gets the </code>AnalysisImage</code>.
	 * @return the </code>AnalysisImage</code>
	 */
	public synchronized AnalysisImage getAImage() {
		return aImage;
	}
	
	/**
	 * Returns the rectangle describing the region of interest.
	 * 
	 * @return the region of intereset
	 */
	public synchronized Rectangle getROI() {
		return new Rectangle(roiX,roiY,roiW,roiH);
	}
	
	/**
	 * Returns the rectangle describing the region of interest used for
	 * threshold calculation.
	 * 
	 * @return the threshold region of interest 
	 */
	public synchronized Rectangle getThresholdROI() {
		return new Rectangle(roi2X,roi2Y,roi2W,roi2H);
	}
	
	/**
	 * Gets the threshold value.
	 * 
	 * @see #setThreshold(double)
	 * 
	 * @return the threshold value
	 */
	public synchronized double getThreshold() {
		return threshold;
	}
	
	/**
	 * Sets the threshold value, which discards all pixels which have
	 * returned by <code>ColorDecoder</code> lover than threshold value. 
	 * Threshold value has effect only between values Const2D and Amplitude2D 
	 * as described by <code>AImage</code>.
	 * 
	 * @see AImage
	 * @see AImage#getAmplitude2D()
	 * @see AImage#getConst2D()
	 * 
	 * @param threshold the threshold value to set
	 */
	public synchronized void setThreshold(final double threshold) {
		if (this.threshold == threshold) return;
//		this.threshold = threshold;
		TineHandler.logger.log(Level.INFO, "Threshold: " + threshold);
		setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Sets the calculate thershold parameter. When true the threshold values
	 * is calculated at each image analysis performed and the original user
	 * defined threshold is overridden with the calculated one.
	 * 
	 * @param calculateThreshold true if the threshold should be calculated or false
	 * 			otherwise
	 */
	public synchronized void setCalculateThreshold(final boolean calculateThreshold) {
		if (this.calculateThreshold == calculateThreshold) return;
		TineHandler.logger.log(Level.INFO, "Calculate threshold: " + calculateThreshold);
		setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Returns true if the threshold is calculated by this engine, or false if
	 * the user defined the threshold to be used.
	 * 
	 * @return true if threshold calculation is automatic
	 */
	public boolean isCalculateThreshold() {
		return calculateThreshold;
	}
	
	/**
	 * Sets the perform fit parameter. When true fitting is performed on the data.
	 * When true, the time required to perform the analysis might be significantly
	 * longer
	 * 
	 * @param performFit true if fitting should be performed or false otherwise
	 */
	public synchronized void setPerformFit(final boolean performFit) {
		if (this.performFit == performFit) return;
		TineHandler.logger.log(Level.INFO, "Perform fit: " + performFit);
		setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Sets the perform smoothing parameter. When true smoothing is performed on the side
	 * view data. 
	 * 
	 * @param performSmoothing true if smoothing should be performed or false otherwise
	 */
	public synchronized void setPerformSmoothing(final boolean performSmoothing) {
		if (this.performSmoothing == performSmoothing) return;
		TineHandler.logger.log(Level.INFO, "Perform smoothing: " + performSmoothing);
		setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,performSmoothing);
	}
	
	/**
	 * Returns true if fitting is performed by this engine, or false if
	 * otherwise
	 * 
	 * @return true if fitting is performed
	 */
	public boolean isPerformFit() {
		return performFit;
	}
	
	/**
	 * Gets the decoder used by this </code>ImageAnalysisEngine</code>.
	 * @return the </code>ColorDecoder</code> used
	 */
	public synchronized ColorDecoder getDecoder() {
		return decoder;
	}
	
	/**
	 * Sets a </code>ColorDecoder</code> to this </code>ImageAnalysisEngine</code>
	 * @param decoder the </code>ColorDecoder</code> to set
	 */
	public synchronized void setDecoder(ColorDecoder decoder) {
		if (this.decoder == decoder) return;
		this.decoder = decoder;
		if (localAnalysis) performAnalysis();
		else {
			// set to server?
		}
	}
	
	/**
	 * Sets the analysis parameters (only region of interest, threshold remains the 
	 * same) and reanalyzes pixels.
	 * @param roiX the starting x coordinate
	 * @param roiY the starting y coordinate
	 * @param roiW the width of the region
	 * @param roiH the height of the region
	 * @param calculateThreshold true if the threshold is calculated by the server or false otherwise
	 * @param performFit true if the fitting should be performed or false otherwise
	 */
	public void setAnalysisParameters(final int roiX, final int roiY, final int roiW, final int roiH, 
			final int roi2X, final int roi2Y, final int roi2W, final int roi2H, 
			final boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
		setAnalysisParameters(threshold, roiX, roiY, roiW, roiH, roi2X, 
				roi2Y, roi2W, roi2H, calculateThreshold,performFit, performSmoothing);
	}
	
	/**
	 * Sets the analysis parameters (threshold and region of interest) and reanalyzes pixels.
	 * @param threshold the threshold
	 * @param roiX the starting x coordinate
	 * @param roiY the starting y coordinate
	 * @param roiW the width of the region
	 * @param roiH the height of the region
	 */
	public synchronized void setAnalysisParameters(final double threshold, 
			final int roiX,	final int roiY, final int roiW, final int roiH, 
			final int roi2X, final int roi2Y, final int roi2W, final int roi2H,
			final boolean calculateThreshold, final boolean performFit,final boolean performSmoothing) {
		
		if (!localAnalysis) {
			executeTask(new Runnable() {
				@Override
				public void run() {
					try {
						getVideoServerConnection().setAnalysisParameters(threshold, roiX, roiY, roiW, roiH, roi2X, roi2Y, roi2W, roi2H,calculateThreshold,performFit);
					} catch (ConnectionFailed e) {
						e.printStackTrace();
					}
				}
			});
		} else {
			this.calculateThreshold = calculateThreshold;
			this.performFit = performFit;
			this.performSmoothing = performSmoothing;
			if (this.threshold == threshold && this.roiX == roiX && this.roiY == roiY && this.roiW == roiW && this.roiH == roiH
					&& this.roi2X == roi2X && this.roi2Y == roi2Y && this.roi2H == roi2H && this.roi2W == roi2W) return;
			this.threshold = threshold;
			if (roiX >= imageWidth) this.roiX = imageWidth-1;
			else if (roiX < 0) this.roiX = 0;
			else this.roiX = roiX;
			if (roiY >= imageHeight) this.roiY = imageHeight-1;
			else if (roiY < 0) this.roiY = 0;
			else this.roiY = roiY;
			if (this.roiX+roiW > imageWidth) this.roiW = imageWidth-this.roiX;
			else if (roiW <= 0) this.roiW = 1;
			else this.roiW = roiW;
			if (this.roiY+roiH > imageHeight) this.roiH = imageHeight-this.roiY;
			else if (roiH <= 0) this.roiH = 1;
			else this.roiH = roiH;

			if (roi2X >= imageWidth) this.roi2X = imageWidth-1;
			else if (roi2X < 0) this.roi2X = 0;
			else this.roi2X = roi2X;
			if (roi2Y >= imageHeight) this.roi2Y = imageHeight-1;
			else if (roi2Y < 0) this.roi2Y = 0;
			else this.roi2Y = roi2Y;
			if (this.roi2X+roi2W > imageWidth) this.roi2W = imageWidth-this.roi2X;
			else if (roi2W <= 0) this.roi2W = 1;
			else this.roi2W = roi2W;
			if (this.roi2Y+roi2H > imageHeight) this.roi2H = imageHeight-this.roi2Y;
			else if (roi2H <= 0) this.roi2H = 1;
			else this.roi2H = roi2H;
			performAnalysis();
		}
		
	}
	
	/**
	 * Sets the analysis mode for this </code>ImageAnalysisEngine</code>.
	 * @param local </code>true</code> for local analysis and
	 * </code>false</code> for remote analysis
	 */
	public void setLocalAnalysisMode(boolean local) {
		if (localAnalysis == local) return;
		TineHandler.logger.log(Level.INFO, "Analysis mode: " + local);
		localAnalysis = local;
		if (localAnalysis) {
			executeTask(new Runnable() {
				@Override
				public void run() {
					getVideoServerConnection().disconnect();
				}
			});
			setRoi(0, 0, imageWidth, imageHeight);
		}
	}
	
	/**
	 * Gets the analysis mode for this </code>ImageAnalysisEngine</code>.
	 * @return the </code>true</code> if local analysis is selected and
	 * </code>false</code> if remote analysis is selected
	 */
	public boolean isLocalAnalysisMode() {
		return localAnalysis;
	}
	
	/**
	 * Starts connection to remote video analysis server.
	 */
	public void startRemoteAnalysis() {
		TineHandler.logger.log(Level.INFO, "Start remote analysis.");
		try {
			getVideoServerConnection().connect();
		} catch (ConnectionFailed e) {
			e.printStackTrace();
		}
	}
	
//	/**
//	 * Makes a single image update.
//	 */
//	public void updateOnce() {
//		TineHandler.logger.log(Level.INFO, "Executes single remote analysis update.");
//		try {
//			getVideoServerConnection().connect();
//		} catch (ConnectionFailed e) {
//			e.printStackTrace();
//		}
//	}
	
	/**
	 * Stops connection to remote video analysis server.
	 */
	public void stopRemoteAnalysis() {
//		TineHandler.logger.log(Level.INFO, "Stop remote analysis.");
		getVideoServerConnection().disconnect();
	}
	
	/**
	 * Sets </code>ConnectionParameters</code> for remote video analysis server.
	 * @param cp the </code>ConnectionParameters</code> to set
	 */
	public void setAnalysisServerConnectionParameters(ConnectionParameters cp) {
		getVideoServerConnection().setConnectionParameters(cp);
	}
	
	private VideoServerConnection getVideoServerConnection() {
		if (videoServerConnection == null) {
			videoServerConnection = new VideoServerConnection();
			videoServerConnection.addPropertyChangeListener(new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (localAnalysis) return;
					if (evt.getPropertyName() == VideoServerConnection.PROPERTY_A_IMAGE) {
						setAImage((AnalysisImage) evt.getNewValue());
					}
					else if (evt.getPropertyName() == VideoServerConnection.PROPERTY_BACKGROUND) {
						setBackground((IMAGE) evt.getNewValue(), false);
					}
					else if (evt.getPropertyName() == VideoServerConnection.PROPERTY_PRECISE_BACKGROUND) {
						//TODO handle this
//						setPreciseBackground((double[]) evt.getNewValue(), videoServerConnection.getPreciseBackgroundWidth(),false);
					}
				}
			});
		}
		return videoServerConnection;
	}
	
	public synchronized void setAImage(AnalysisImage aImage) {
		this.aImage = aImage;
		pcSupport.firePropertyChange(PROPERTY_A_IMAGE, null, aImage);
	}
	
	/**
	 * Sets the image analyzer which is used for the analysis.
	 * @param imageAnalyzer
	 */
	public void setImageAnalyzer(ImageAnalyzer imageAnalyzer) {
		if (imageAnalyzer == null) {
			imageAnalyzer = new GaussianImageAnalyzer();
		}
		this.imageAnalyzer = imageAnalyzer;
		performAnalysis();
	}
	
	
	/**
	 * Sets the color map to use for the analysis.
	 * 
	 * @param colorMap
	 */
	public void setColorMap(ColorMap colorMap) {
		this.colorMap = colorMap;
		performAnalysis();
	}
	
	private synchronized void performAnalysis() {
		if (image == null) return;
		executeAnalysis(new Runnable(){
			@Override
			public void run() {
				setAImage(imageAnalyzer.performAnalysis(colorMap,decoder,image,preciseBackground,imageWidth,imageHeight,roiX,roiY,roiW,roiH,
						roi2X, roi2Y, roi2W, roi2H, threshold,calculateThreshold,performFit,performSmoothing));
				
			}
		});
	}
			
	protected void executeAnalysis(Runnable r) {
		if (analysisExecutor == null) {
			ThreadFactory factory = new TINEThreadFactory("TINE-ImageAnalysis");
			analysisExecutor = new  ThreadPoolExecutor(1, 1, 0, TimeUnit.NANOSECONDS,
		              new DismissableBlockingQueue<Runnable>(2, true), factory){
				@Override
				protected void afterExecute(Runnable r, Throwable t) {
					super.afterExecute(r, t);
					if (t != null) {
						t.printStackTrace();
					}
				}
			};
			analysisExecutor.setRejectedExecutionHandler(new RejectedExecutionHandler(){
				public void rejectedExecution(Runnable r, ThreadPoolExecutor executor) {
					//ignore
				}
			});
		}

		analysisExecutor.execute(r);
	}
	
	@Deprecated
	@SuppressWarnings("unused")
	private static int tineFormatToBufferedType(IMAGE im) {
		int format = im.getFrameHeader().imageFormat;
		if (format == ImageFormat.IMAGE_FORMAT_GRAY.getId()) {
			return BufferedImage.TYPE_BYTE_GRAY;
		} else if (format == ImageFormat.IMAGE_FORMAT_RGB.getId()) {
			return BufferedImage.TYPE_INT_RGB;
		}
		return BufferedImage.TYPE_INT_RGB;
	}
		
	private void executeTask(Runnable r) {
		if (executor == null) {
			executor = Executors.newSingleThreadExecutor();
		}
		executor.execute(r);
	}
		
	/**
	 * Adds a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		pcSupport.addPropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		pcSupport.removePropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Adds a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener) {
		pcSupport.addPropertyChangeListener(listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener) {
		pcSupport.removePropertyChangeListener(listener);
	}
	
	/**
	 * Transforms the IMAGE to BufferedImage using the appropriate algorithm
	 * for the given image.
	 * 
	 * @param image the image to transform
	 * @return transformed image
	 */
	public static BufferedImage toBufferedImage(IMAGE image, ColorMap colorMap) {
		if (TBufferedImage.isSupported(image.getFrameHeader().imageFormat)) {
			return TBufferedImage.toBufferedImage(image,colorMap,false,false);
		} else {
			return new TImageConverter().parseTineImage(image,false);
		}
	}
}

