package de.desy.acop.video.analysis.interferometer;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.tine.types.IMAGE;

/**
 * 
 * <code>IFMImage</code> is the object that is transferred across the
 * system from the video analysis server to the video client.
 * This object carries with it the raw tine IMAGE and analysis results
 * and parameters. For details on parameters see the constructor
 * documentation.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class IFMImage extends AnalysisImage {

	private final double[] constantY = new double[1];
	private final double[] amplitudeY = new double[1];
	private final double[] asincY = new double[1];
	private final double[] psincY = new double[1];
	private final double[] acosY = new double[1];
	private final double[] pcosY = new double[1];
	private final double[] visY = new double[1];
	private final byte[] vertical = new byte[1]; 
	
//	private final double[] sideViewAmplitudeX = new double[1];
//	private final double[] sideViewConstX = new double[1];
//	private final double[] meanX = new double[1];
//	private final double[] meanXFit = new double[1];
//	private final double[] stdXFit = new double[1];
//	private final double[] amplitudeXFit = new double[1];
//	private final double[] constXFit = new double[1];
//	private final double[] slopeXFit = new double[1];
	
	
	/**
	 * Constructs a new IFMImage, with default values (empty image and 0 value parameters
	 * and results).
	 */
	public IFMImage() {
		super("IFMISr4");
		initStructure();
	}
	
	/**
	 * Constructs a new IFMImage from the given IFMImage. This method creates a clone
	 * of the given image.
	 * 
	 * @param image the image
	 */
	public IFMImage(IFMImage image) {
		this(image.getImage(),
				image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),
				image.getRoi2X(), image.getRoi2Y(), image.getRoi2W(), image.getRoi2H(), image.getThreshold(),
				image.getSideViewX(), image.getSideViewY(), 
				image.getConstantY(), image.getAmplitudeY(), image.getAsincY(), image.getPsincY(),
				image.getAcosY(), image.getPcosY(), image.getVisY(),
				image.getThresholdPoints(),image.isCalculateThreshold(),
				image.isPerformFit(),image.isPerformSmoothing(),
				image.isVertical());
	}
	
	/**
	 * Constructs a new IFMImage.
	 *  
	 * @param image the raw image, which was analyzed and all data was obtained from
	 * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
	 * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
	 * @param roiW the width of region of interest (image is analyzed inside region of interest)
	 * @param roiH the height of region of interest (image is analyzed inside region of interest)
	 * @param roi2X the region of interest used for automatic threshold calculation right coordinate
	 * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
	 * @param roi2W the region of interest used for automatic threshold calculation width
	 * @param roi2H the region of interest used for automatic threshold calculation height
	 * @param threshold the used threshold value (calculated or user defined)
	 * @param sideViewX horizontal side view values
	 * @param sideViewY vertical side view values
	 * @param constantY the background constant of the Y fit
	 * @param amplitudeY the amplitude of the Y fit
	 * @param asincY the asinc parameter of the Y fit
	 * @param psincY the psinc parameter of the Y fit
	 * @param acosY the acos parameter of the Y fit
	 * @param pcosY the pcos parameter of the Y fit
	 * @param visY the visibility parameter of the Y fit
	 * @param thresholdPoints the number of points on which the threshold was calculated
	 * @param calculateThreshold flag indicating if threshold calculation is automatic
	 * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters
	 * 			are a-priori invalid)
	 * @param performSmoothing flag indicating if smoothing should be performed
	 * @param vertical true if this image represents vertically oriented interferometer or false
	 * 			if it represents horizontally oriented interferometer
	 */
	public IFMImage(IMAGE image, 
			int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y,int roi2W, int roi2H, double threshold,
			double[] sideViewX, double[] sideViewY,  
			double constantY, double amplitudeY, double asincY, double psincY, 
			double visY, double acosY, double pcosY,
			int thresholdPoints, boolean calculateThreshold, boolean performFit, boolean performSmoothing,
			boolean vertical) {
		super("IFMISr4",image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,
				sideViewX,sideViewY,0,calculateThreshold,performFit,performSmoothing);

		this.constantY[0] = constantY;
		this.amplitudeY[0] = amplitudeY;
		this.pcosY[0] = pcosY;
		this.acosY[0] = acosY;
		this.asincY[0] = asincY;
		this.psincY[0] = psincY;
		this.visY[0] = visY;
		this.vertical[0] = vertical ? (byte)1 : (byte)0;
		
		initStructure();
		initDone();
	}
	
	private void initStructure() {
		addField(constantY,"constantY");
		addField(amplitudeY,"amplitudeY");
	    addField(asincY,"asincY");
		addField(psincY,"psincY");
		addField(acosY,"acosY");
		addField(pcosY,"pcosY");
	    addField(visY,"visY");
	    addField(vertical,"vertical");
	}
	
	/**
	 * @return true if this image is vertical or false if horizontal
	 */
	public boolean isVertical() {
		return vertical[0] == 1;
	}

	/**
	 * @return 1 if the image is vertical or 0 if horizontal
	 */
	public byte[] getVertical() {
		return vertical;
	}
	
	
	/**
	 * @return the constantY
	 */
	public double getConstantY() {
		return constantY[0];
	}

	
	/**
	 * @return the amplitudeY
	 */
	public double getAmplitudeY() {
		return amplitudeY[0];
	}

	
	/**
	 * @return the asincY
	 */
	public double getAsincY() {
		return asincY[0];
	}

	
	/**
	 * @return the psincY
	 */
	public double getPsincY() {
		return psincY[0];
	}

	
	/**
	 * @return the acosY
	 */
	public double getAcosY() {
		return acosY[0];
	}

	
	/**
	 * @return the pcosY
	 */
	public double getPcosY() {
		return pcosY[0];
	}

	
	/**
	 * @return the visY
	 */
	public double getVisY() {
		return visY[0];
	}	
}
