package de.desy.acop.video.analysis.interferometer;

import java.awt.GridLayout;
import java.awt.image.BufferedImage;

import javax.swing.JFrame;
import javax.swing.JPanel;

import de.desy.acop.chart.Acop;
import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.ColorDecoder;
import de.desy.acop.video.analysis.FittingUtilities;
import de.desy.acop.video.analysis.FittingUtilities.InterferometerCostFunction;
import de.desy.acop.video.analysis.ImageAnalysisEngine;
import de.desy.acop.video.analysis.ImageAnalyzer;
import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;

/**
 * 
 * <code>InterferometerImageAnalyzer</code> fits the interferometer function to the vertical side view.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class InterferometerImageAnalyzer implements ImageAnalyzer {
	
	private static double[] NAN_ARRAY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
	private double[] previousStartValuesIf = null;
	
	private static final double COST_MIN = 3;	
	
	private final boolean verticalInterferometer;
	
	public InterferometerImageAnalyzer(boolean verticalInterferometer) {
		this.verticalInterferometer = verticalInterferometer;
	}
	
	@Override
	public AnalysisImage performAnalysis(ColorMap colorMap, ColorDecoder decoder, IMAGE im, double[] bg, int iW, int iH, int rX,
			int rY, int rW,	int rH, int r2X, int r2Y, int r2W, int r2H, double threshold, 
			boolean calculateThreshold, boolean performFit,	boolean smoothing) {
		if (verticalInterferometer) {
			return performAnalysisVertical(decoder,im,bg,iW,iH,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,calculateThreshold,performFit,smoothing);
		} else {
			return performAnalysisHorizontal(decoder,im,bg,iW,iH,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,calculateThreshold,performFit,smoothing);
		}
		
	}
	
	private AnalysisImage performAnalysisVertical(ColorDecoder decoder, IMAGE im, double[] bg, int iW, int iH, int rX,
			int rY, int rW,	int rH, int r2X, int r2Y, int r2W, int r2H, double threshold, 
			boolean calculateThreshold, boolean performFit,	boolean smoothing) {
		IMAGE newImage = im;
		BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(im,ColorMap.GRAYSCALE);
		int aH = im.getFrameHeader().aoiHeight;
		int aW = im.getFrameHeader().aoiWidth;

		if (aH != -1) {
			iW = aW;
			iH = aH;
		}
		
		if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
		
		int[] data = bufferedImage.getRGB(0, 0, iW, iH, null, 0, iW);
		double[] values = decoder.transform(data);
		
		int i, j;
		
		if (bg != null) {		
			int[] newImageVals = new int[values.length];
			try { 
				i = 0;
				//try catch is actually faster than anything else
    			for (;;) {
    				values[i] -= bg[i];
    				newImageVals[i] = (int)(values[i]+0.5);
    				if (newImageVals[i] < 0) {
    					newImageVals[i] = 0;
    				}
    				i++;
    			}
			} catch (ArrayIndexOutOfBoundsException e) {
				//ignore
			}
			bufferedImage = new BufferedImage(bufferedImage.getWidth(),bufferedImage.getHeight(),BufferedImage.TYPE_BYTE_GRAY);
			bufferedImage.setRGB(0,0,iW,iH,newImageVals,0,iW);
			newImage = TBufferedImage.toIMAGE(bufferedImage);
			newImage.getFrameHeader().aoiWidth = im.getFrameHeader().aoiWidth;
			newImage.getFrameHeader().aoiHeight = im.getFrameHeader().aoiHeight;
			newImage.getFrameHeader().xStart = im.getFrameHeader().xStart;
			newImage.getFrameHeader().yStart = im.getFrameHeader().yStart;
			newImage.getFrameHeader().sourceHeight = im.getFrameHeader().sourceHeight;
			newImage.getFrameHeader().sourceWidth = im.getFrameHeader().sourceWidth;
		}
		
		//calculate the threshold value
		int thresholdPointsCount = 0;
		if (calculateThreshold) {
			double sum = 0;
			int r = rX + rW, b = rY+rH;
			boolean inside = false;
			for(j = 0; j < r2H; j++)	{
				inside = rY <= r2Y + j && r2Y + j < b;
				for(i = 0; i < r2W; i++) {
					//do not include indexes, which are in normal ROI
					if (rX <= r2X + i && r2X + i < r && inside) continue;
					sum += values[i+r2X+iW*(j+r2Y)];
					thresholdPointsCount++;
				}
			}
			if (thresholdPointsCount != 0) {
				threshold = sum/thresholdPointsCount;
			} else {
				threshold = Double.NaN;
			}
		}
		double[] sideViewX = new double [iW];
		double[] sideViewY = new double [iH];
		
		double maxValue = -Double.MAX_VALUE;
		double minValue = Double.MAX_VALUE;
		double value;
		if (Double.isNaN(threshold)) {
			for (i = 0; i < iW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (maxValue < value) maxValue = value;
    				sideViewX[i] += (value);
    			}
    		}
		} else {
    		for (i = 0; i < iW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (value < threshold) continue;
    				if (maxValue < value) maxValue = value;
    				sideViewX[i] += (value-threshold);
    			}
    		}
		}
		
		//smoothing 	
		if (smoothing) {
			sideViewX = FittingUtilities.smooth(sideViewX);
		}
		//end of smoothing
					
		try {
			i = 0;
			for (;;) {
				if (values[i] < 0) values[i] = 0;
				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			//ignore
		}
		
		
		double[] xMinMax = findMinMax(sideViewX,0,iW);
		
		rX = (int)xMinMax[2]-5;
		if (rX < 0) rX = 1;
		rW = 10; 
		double[] roiSideViewY = new double[rH];
		if (Double.isNaN(threshold)) {
			for (i = rX; i < rX+rW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (maxValue < value) maxValue = value;
    				sideViewY[j] += (value);
    			}
    		}
		} else {
    		for (i = rX; i < rX+rW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (value < threshold) continue;
    				if (maxValue < value) maxValue = value;
    				sideViewY[j] += (value-threshold);
    			}
    		}
		}
		try {
			i = 0;
			for (;;) {
				sideViewY[i] = sideViewY[i]/rW;
				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			//ignore
		}
		
		System.arraycopy(sideViewY,rY,roiSideViewY,0,rH);
		double[] startValuesY;
		
		if (performFit) {
    		try {       
    			startValuesY = tryFit(roiSideViewY);
    			if (startValuesY != null) {
    				previousStartValuesIf = startValuesY;
    			}
    		} catch (Exception e) {
    			startValuesY = NAN_ARRAY;
    			previousStartValuesIf = null;
    		}
    		
    		if (FittingUtilities.isNaN(startValuesY)) {
    			startValuesY = NAN_ARRAY;
    			previousStartValuesIf = null;
    		}
		} else {
			startValuesY = NAN_ARRAY;
			previousStartValuesIf = null;
		}
		

		try {
			if (!Double.isNaN(threshold)){
				maxValue = maxValue - threshold; 
			}
			IFMImage newAImage = new IFMImage(
					newImage, rX, rY, rW, rH, r2X, r2Y, r2W, r2H, threshold,
					sideViewX, sideViewY,
					startValuesY[0], startValuesY[1],startValuesY[2],startValuesY[3],startValuesY[4],
					startValuesY[5], startValuesY[6], 
					thresholdPointsCount,calculateThreshold,performFit,smoothing,true
			);
			return newAImage;				
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	private double[] tryFit(double[] sideView) throws Exception {
		int l = sideView.length;
		double[] steps = new double[l];
		double[] weights = new double[l];
		for (int i = 0; i < l; i++) {
			steps[i] = i;
			weights[i] = 1;
		}

		double cost = 0;
		double[] startValues;
		FittingUtilities.InterferometerCostFunction f = new InterferometerCostFunction(steps,sideView);
		//first try to use the previous values as a strating point
		if (previousStartValuesIf != null) {
			startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,previousStartValuesIf);
			cost = Math.sqrt(f.value(startValues)/l);
			if (cost < COST_MIN && !isResultInvalid(startValues)) {
				return startValues;
			}
			//if levenberg fails, try nelder mead
			startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,previousStartValuesIf);
			cost = Math.sqrt(f.value(startValues)/l);
			if (cost < COST_MIN && !isResultInvalid(startValues)) {
				return startValues;
			}
		}
		
		//estimates start values and try levenberg
		double[] freshStart = getStartValues(sideView,false); 
		startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,freshStart);
		cost = Math.sqrt(f.value(startValues)/l);
		if (cost < COST_MIN && !isResultInvalid(startValues)) {
			return startValues;
		} 	
		//try nelder mead with estimated start values
		startValues = freshStart;
		startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,freshStart);
		cost = Math.sqrt(f.value(startValues)/l);
		if (cost < COST_MIN && !isResultInvalid(startValues)) {
			return startValues;
		} 	

		//randomize start values and try levenberg
		freshStart = getStartValues(sideView,true);
		startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,freshStart);
		cost = Math.sqrt(f.value(startValues)/l);
		if (cost < COST_MIN && !isResultInvalid(startValues)) {
			return startValues;
		} 
		//randomize start values and try nelder mead
		freshStart = getStartValues(sideView,true);
		startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,freshStart);
		cost = Math.sqrt(f.value(startValues)/l);
		if (cost < COST_MIN && !isResultInvalid(startValues)) {
			return startValues;
		} 
	
		//if nothing converged return null
		return null;
	}
		
	private AnalysisImage performAnalysisHorizontal(ColorDecoder decoder, IMAGE im, double[] bg, int iW, int iH, int rX,
			int rY, int rW,	int rH, int r2X, int r2Y, int r2W, int r2H, double threshold, 
			boolean calculateThreshold, boolean performFit,	boolean smoothing) {
		
		IMAGE newImage = im;
		BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(im,ColorMap.GRAYSCALE);
		int aH = im.getFrameHeader().aoiHeight;
		int aW = im.getFrameHeader().aoiWidth;

		if (aH != -1) {
			iW = aW;
			iH = aH;
		}
		
		if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
		
		int[] data = bufferedImage.getRGB(0, 0, iW, iH, null, 0, iW);
		double[] values = decoder.transform(data);
		
		int i, j;
		
		if (bg != null) {		
			int[] newImageVals = new int[values.length];
			try { 
				i = 0;
				//try catch is actually faster than anything else
    			for (;;) {
    				values[i] -= bg[i];
    				newImageVals[i] = (int)(values[i]+0.5);
    				if (newImageVals[i] < 0) {
    					newImageVals[i] = 0;
    				}
    				i++;
    			}
			} catch (ArrayIndexOutOfBoundsException e) {
				//ignore
			}
			bufferedImage = new BufferedImage(bufferedImage.getWidth(),bufferedImage.getHeight(),BufferedImage.TYPE_BYTE_GRAY);
			bufferedImage.setRGB(0,0,iW,iH,newImageVals,0,iW);
			newImage = TBufferedImage.toIMAGE(bufferedImage);
			newImage.getFrameHeader().aoiWidth = im.getFrameHeader().aoiWidth;
			newImage.getFrameHeader().aoiHeight = im.getFrameHeader().aoiHeight;
			newImage.getFrameHeader().xStart = im.getFrameHeader().xStart;
			newImage.getFrameHeader().yStart = im.getFrameHeader().yStart;
			newImage.getFrameHeader().sourceHeight = im.getFrameHeader().sourceHeight;
			newImage.getFrameHeader().sourceWidth = im.getFrameHeader().sourceWidth;
		}
		
		//calculate the threshold value
		int thresholdPointsCount = 0;
		if (calculateThreshold) {
			double sum = 0;
			int r = rX + rW, b = rY+rH;
			boolean inside = false;
			for(j = 0; j < r2H; j++)	{
				inside = rY <= r2Y + j && r2Y + j < b;
				for(i = 0; i < r2W; i++) {
					//do not include indexes, which are in normal ROI
					if (rX <= r2X + i && r2X + i < r && inside) continue;
					sum += values[i+r2X+iW*(j+r2Y)];
					thresholdPointsCount++;
				}
			}
			if (thresholdPointsCount != 0) {
				threshold = sum/thresholdPointsCount;
			} else {
				threshold = Double.NaN;
			}
		}
		double[] sideViewX = new double [iW];
		double[] sideViewY = new double [iH];
		
		double maxValue = -Double.MAX_VALUE;
		double minValue = Double.MAX_VALUE;
		double value;
		if (Double.isNaN(threshold)) {
			for (i = 0; i < iW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (maxValue < value) maxValue = value;
    				sideViewY[j] += (value);
    			}
    		}
		} else {
    		for (i = 0; i < iW; ++i) {
    			for (j = 0; j < iH; ++j) {
    				value = values[i+j*iW];
    				if (minValue > value) minValue = value;
    				if (value < threshold) continue;
    				if (maxValue < value) maxValue = value;
    				sideViewY[j] += (value-threshold);
    			}
    		}
		}
		
		//smoothing 	
		if (smoothing) {
			sideViewY = FittingUtilities.smooth(sideViewY);
		}
		//end of smoothing

					
		try {
			i = 0;
			for (;;) {
				if (values[i] < 0) values[i] = 0;
				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			//ignore
		}
		
		
		double[] yMinMax = findMinMax(sideViewY,0,iH);
		
		rY = (int)yMinMax[2]-5;
		if (rY < 0) rY = 1;
		rH = 10; 
		double[] roiSideViewX = new double[rW];
		if (Double.isNaN(threshold)) {
			for (i = rY; i < rY+rH; ++i) {
    			for (j = 0; j < iW; ++j) {
    				value = values[i*iW+j];
    				if (minValue > value) minValue = value;
    				if (maxValue < value) maxValue = value;
    				sideViewX[j] += (value);
    			}
    		}
		} else {
    		for (i = rY; i < rY+rH; ++i) {
    			for (j = 0; j < iW; ++j) {
    				value = values[i*iW+j];
    				if (minValue > value) minValue = value;
    				if (value < threshold) continue;
    				if (maxValue < value) maxValue = value;
    				sideViewX[j] += (value-threshold);
    			}
    		}
		}
		try {
			i = 0;
			for (;;) {
				sideViewX[i] = sideViewX[i]/rH;
				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			//ignore
		}
		
		System.arraycopy(sideViewX,rX,roiSideViewX,0,rW);
		double[] startValuesX;
		
		if (performFit) {
    		try {       
    			startValuesX = tryFit(roiSideViewX);
    			if (startValuesX != null) {
    				previousStartValuesIf = startValuesX;
    			}
    		} catch (Exception e) {
    			startValuesX = NAN_ARRAY;
    			previousStartValuesIf = null;
    		}
    		
    		if (FittingUtilities.isNaN(startValuesX)) {
    			startValuesX = NAN_ARRAY;
    			previousStartValuesIf = null;
    		}
		} else {
			startValuesX = NAN_ARRAY;
			previousStartValuesIf = null;
		}
		
		try {
			if (!Double.isNaN(threshold)){
				maxValue = maxValue - threshold; 
			}
			IFMImage newAImage = new IFMImage(
					newImage, rX, rY, rW, rH, r2X, r2Y, r2W, r2H, threshold,
					sideViewX, sideViewY,
					startValuesX[0], startValuesX[1],startValuesX[2],startValuesX[3],startValuesX[4],
					startValuesX[5], startValuesX[6], 
					thresholdPointsCount,calculateThreshold,performFit,smoothing,false
			);
			return newAImage;				
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Checks if the result is invalid. Result is invalid if the visibility is less than 0 or
	 * if any of the elements is too large.
	 * 
	 * @param result the result to check
	 * @return true if the result is invalid or false if valid
	 */
	private static boolean isResultInvalid(double[] result) {
		if (result[4] < 0) {
			result[4] = -result[4];
			result[6] = result[6] + Math.PI/result[5];
		}
		for (int i = 0; i < result.length; i++) {
			if (Math.abs(result[i]) > 1E10) return true;
		}
		return false;
	}
	
	/**
	 * Estimates the starting values for the fit.
	 * @param sideView the profile which the fit is calculated for
	 * @return the starting values
	 */
	private static double[] getStartValues(double[] sideView, boolean randomize) {
		double[] yMinMax = findMinMax(sideView,0,sideView.length);
		double fr = getFrequency(sideView);
		double bcg = Math.max(yMinMax[0],Math.random()*3);
		double amp = yMinMax[1]-bcg;
		double asin = 0.014;
		double psin = yMinMax[2];
		double vis = 0.9;
		double acos = fr;//(fr > 0.1 && fr < 0.4) ? fr : 0.3;
		double pcos = 200;
		if (randomize) {
			vis = 0.2 + Math.random() * 0.75;
			asin = Math.random()*0.02;
			pcos = 100 + Math.random();
		}
		return new double[]{bcg,amp,asin,psin,vis,acos,pcos};
	}
	
	/**
	 * Finds the minimum and maximum value in the array looking at the elements between
	 * start and start+length.
	 * @param values the values to check
	 * @param start the starting index
	 * @param length the length of the check
	 * @return the triple min/max/index
	 */
	private static double[] findMinMax(double[] values,int start, int length) {
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		int index = 0;
		for (int i = start; i < start+length; i++) {
			if (values[i] > max) {
				max = values[i];
				index = i;
			}
			if (values[i] < min) min = values[i];
		}
		return new double[] {min, max,index};
	}
	
	/**
	 * Calculates the most important frequency that is found in the given function.
	 * @param y the function
	 * @return the frequency
	 */
	private static double getFrequency(double[] y) {
		double[] f = null;
		int l = y.length;
		if (l <= 64) f = new double[64];
		else if (l <= 128) f = new double[128];
		else if (l <= 256) f = new double[256];
		else if (l <= 512) f = new double[512];
		else if (l <= 1024) f = new double[1024];
		else if (l <= 2048) f = new double[2048];
		else if (l <= 4096) f = new double[4096];
		else f = new double[8192];
		
		int d = f.length;
		System.arraycopy(y,0,f,0,y.length);
		double[][] fft = FittingUtilities.FFT1D(new double[][]{f,new double[d]},true);
		int index = 0;
		double max = 0;
		double t;
		double tm1 = Math.sqrt(fft[0][0]*fft[0][0]+fft[1][0]*fft[1][0]);
		double tp1 = Math.sqrt(fft[0][2]*fft[0][2]+fft[1][2]*fft[1][2]);
		t = Math.sqrt(fft[0][1]*fft[0][1]+fft[1][1]*fft[1][1]);
		for (int i = 10; i < d/2; i++) {
			if (t > max && t > tm1 && t > tp1) {
				max = t;
				index = i-1;
			}
			tm1 = t;
			t = tp1;
			tp1 = Math.sqrt(fft[0][i]*fft[0][i]+fft[1][i]*fft[1][i]);		
		}
				
		return index*2*Math.PI/d;
	}

	
	public static void main(String[] args) throws Exception {
//		double[] a = new double[]{0,100,0.015,250,0.9,0.21,195};
//		double[] y = new double[512];
//		double[] x = new double[y.length];
//		double[] w = new double[y.length];
//		double sinc,xx;
//		double k = 1.;
//		for (int i = 0; i < y.length; i++) {
//			xx = a[2]*(i/k-a[3]);
//    		if (Math.abs(xx) < 1e-16) {
//    			sinc = 1;
//    		} else {
//    			sinc = Math.sin(xx)/xx;
//    		}
//    		y[i] = a[0] + a[1] * sinc*sinc*(1+a[4]*Math.cos(a[5]*(i/k-a[6])));
//    		x[i] = i/k;
//    		w[i] = 1;
//			
//		}
//		
//		double[] ret = FittingUtilities.lmInterferometer(false,x,y,w,new double[]{0.1,90,.015,200,1,0.20,200});
//		System.out.println(Arrays.toString(ret));
//		ret = FittingUtilities.lmInterferometer(false,x,y,w,new double[]{5,40,.01,100,0.,0.22,200});
//		System.out.println(Arrays.toString(ret));
//		for (int b = 0; b < 5; b++) {
//			double bg = Math.random();
//			double amp = Math.random()*100;
//			double asin = Math.random()*0.01+0.01;
//			double psin = Math.random()*200;
//			double acos = 0.22;
//			double pcos = 200;
//			double vis = Math.random()*0.1+0.8;
//			ret = FittingUtilities.lmInterferometer(false,x,y,w,new double[]{bg,amp,asin,psin,vis,acos,pcos});
//			System.out.println(Arrays.toString(ret));
//		}
		
		double omega = 10;
		int n = 1800;
		double[] y = new double[n];
		double[] x = new double[n];
		for (int i = 0; i < n; i++) {
			x[i] = 2*Math.PI*i/n;			
			y[i] = Math.sin(omega*x[i]);
		}
		System.out.println(getFrequency(y));
		
		Acop acop = new Acop();
		acop.setYBestScale(false);
		acop.setXBestScale(false);
		acop.setXMin(0);
		acop.setXMax(7);
		acop.setYMin(-2);
		acop.setYMax(2);
		acop.draw(y,x);
		
		double[] yy = new double[2048];
		System.arraycopy(y,0,yy,0,n);
		
		double[][] fft = FittingUtilities.FFT1D(new double[][]{yy,new double[2048]},true);
		double[] d = new double[2048];
		for (int i = 0; i < n; i++) {
			d[i] = fft[0][i]*fft[0][i] + fft[1][i]*fft[1][i];
		}
		Acop acop2 = new Acop();
		acop2.setYBestScale(false);
		acop2.setXBestScale(false);
		acop2.setXMin(0);
		acop2.setXMax(100);
		acop2.setYMin(0);
		acop2.setYMax(200);
		acop2.draw(d);
		acop2.autoScale(false,false,true,false);
		
		JPanel p = new JPanel(new GridLayout(1,2));
		p.add(acop);
		p.add(acop2);
		
		JFrame f = new JFrame();
		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		f.setContentPane(p);
		f.setSize(800,400);
		f.setVisible(true);
		
	}
}

