package de.desy.tine.addrUtils;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.definitions.TTransport;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

/**
 * @author duval
 *
 * This class hold both the Server address information and the FEC address
 * information and corresponds to ExpDataStruct followed by FecDataStruct
 * in the C libraries.
 *  
 */
public final class SrvAddr
{
  public String fecName;  // = new String(new StringBuffer(16));
  public String subSystem;
  public String expName; // = new String(new StringBuffer(16));
  public String eqmName; // = new String(new StringBuffer(7));
  public String eqmContext; // = new String(new StringBuffer(16));
  public byte[] ipxNetwork = new byte[4];
  public byte[] ipxNode = new byte[6];
  public byte[] ipxImmediateAddress = new byte[6];
  public String ipAddr; // = new String(new StringBuffer(16));
  public int portOffset;
  public byte[] ipHAddr = new byte[4];
  public int inetProtocol;
  public int tineProtocol;
  public static int sizeInBytes =
    TStrings.FEC_NAME_SIZE +
    TStrings.SUBSYSTEM_NAME_SIZE +
    TStrings.EXPORT_NAME_SIZE +
    TStrings.EQM_NAME_SIZE +
    TStrings.CONTEXT_NAME_SIZE +
    FECAddr.sizeInBytes;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  private byte[] hByteArray;
  SrvAddr()
  {
  }
  public SrvAddr(FECAddr fecAddr)
  {
    fecName = fecAddr.fecName;
    ipAddr = fecAddr.ipAddr;
    portOffset = fecAddr.portOffset;
    if (fecAddr.ipHAddr[0] == 0 && fecAddr.ipHAddr[1] == 0 && 
        fecAddr.ipHAddr[2] == 0 && fecAddr.ipHAddr[3] == 0)
    {
      try
      {
        InetAddress ia = InetAddress.getByName(ipAddr.trim());
        fecAddr.ipHAddr = ia.getAddress();
      }
      catch (UnknownHostException e)
      {
        e.printStackTrace();
        MsgLog.log("SrvAddr",e.getMessage(),TErrorList.code_failure,e,0);
      }      
    }
    System.arraycopy(fecAddr.ipHAddr,0,ipHAddr,0,4);
  }
  public SrvAddr(String fecName,String fecIP,int fecPort,String eqmName,String expName,String ctxName)
  {
    this.fecName = new String(fecName);
    eqmContext = new String(ctxName);
    this.expName = new String(expName);
    this.eqmName = new String(eqmName);
    ipAddr = new String(fecIP);
    portOffset = fecPort;
    inetProtocol = TTransport.UDP;
    tineProtocol = TTransport.DEFAULT_PROTOCOL_LEVEL;  
    try
    {
      InetAddress ia = InetAddress.getByName(ipAddr.trim());
      System.arraycopy(ia.getAddress(),0,ipHAddr,0,4);
    }
    catch (UnknownHostException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr",e.getMessage(),TErrorList.code_failure,e,0);
    }      
  }
  public int getPortOffset()
  {
    return portOffset;
  }
  private void pushStringToBytes(String s,byte[] b,int maxlen)
  {
    int i,len;
    if (s != null)
    {
      if((len=s.length()) > maxlen) len = maxlen;
      System.arraycopy(s.getBytes(),0,b,0,len);
    }
    else len = 0;
    for (i=len; i<maxlen; i++) b[i] = 0;    
  }
  public byte[] toByteArray()
  { 
    try
    { // prepare double-address info for export to ENS
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      // First: export name information (ExpDataStruct in c)
      byte[] bstr = new byte[64];
      pushStringToBytes(fecName, bstr, TStrings.FEC_NAME_SIZE);
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      pushStringToBytes(subSystem, bstr, TStrings.SUBSYSTEM_NAME_SIZE);
      ds.write(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE);
      pushStringToBytes(expName, bstr, TStrings.EXPORT_NAME_SIZE);
      ds.write(bstr,0,TStrings.EXPORT_NAME_SIZE);
      pushStringToBytes(eqmName, bstr, TStrings.EQM_NAME_SIZE);
      ds.write(bstr,0,TStrings.EQM_NAME_SIZE);
      pushStringToBytes(eqmContext, bstr, TStrings.CONTEXT_NAME_SIZE);
      ds.write(bstr,0,TStrings.CONTEXT_NAME_SIZE);
      // Second: fec name information (FecDataStruct in c)
      pushStringToBytes(fecName, bstr, TStrings.FEC_NAME_SIZE);
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      ds.write(ipxNetwork,0,4);
      ds.write(ipxNode,0,6);
      ds.write(ipxImmediateAddress,0,6);
      pushStringToBytes(ipAddr, bstr, TStrings.ADDR_SIZE);
      ds.write(bstr,0,TStrings.ADDR_SIZE);
      ds.write(ipHAddr,0,4);
      ds.writeInt(Swap.Long(portOffset));
      ds.writeInt(Swap.Long(inetProtocol));
      ds.writeInt(Swap.Long(tineProtocol));
      ds.close();	
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }	 
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      // First: export information:
      byte[] bstr = new byte[64];
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr,0,TStrings.FEC_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE);
      subSystem = new String(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.EXPORT_NAME_SIZE);
      expName = new String(bstr,0,TStrings.EXPORT_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.EQM_NAME_SIZE);
      eqmName = new String(bstr,0,TStrings.EQM_NAME_SHORTSIZE).trim();
      ds.read(bstr,0,TStrings.CONTEXT_NAME_SIZE);
      eqmContext = new String(bstr,0,TStrings.CONTEXT_NAME_SHORTSIZE).trim();
      // Second: FEC address information:
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr,0,TStrings.FEC_NAME_SIZE).trim();
      ds.read(ipxNetwork,0,4);
      ds.read(ipxNode,0,6);
      ds.read(ipxImmediateAddress,0,6);
      ds.read(bstr,0,TStrings.ADDR_SIZE);
      ipAddr = new String(bstr,0,TStrings.ADDR_SIZE).trim();
      ds.read(ipHAddr,0,4);
      portOffset = Swap.Long(ds.readInt());
      inetProtocol = Swap.Long(ds.readInt());
      tineProtocol = Swap.Long(ds.readInt());
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
    }	 
  }  
  public String toString()
  {
    String msg = new String("Exp name : " + expName);
    msg = msg.concat("\nEqm name : " + eqmName);
    msg = msg.concat("\nContext  : " + eqmContext);
    msg = msg.concat("\nFEC name : " + fecName);
    msg = msg.concat("\nFEC ip   : " + ipAddr);
    msg = msg.concat("\nFEC portOffset : " + portOffset);
    msg = msg.concat("\ntine protocol : " + tineProtocol);
    return msg;  
  }
}