/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package de.desy.tine.structUtils;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.types.*;
import java.util.*;
import java.io.*;

/**
 * Tagged Structure The transfer of 'user-defined' data structures is
 *        allowed in TINE. In order to avoid nasty problems concerning alignment
 *        and byte swapping it is strongly suggested to extend the
 *        TTaggedStructure class when creating a data structure.
 */
public class TTaggedStructure implements Cloneable
{
  private String structName;
  private boolean registered = false;
  private boolean hasExtendedSpace = false;
  private int sizeInBytes;
  private int extendedOffset;
  private int extendedSizeInBytes = 0;
  private ArrayList<structField> fieldArrayList = null;
  static private int indent = 0; // used by toString()
  public boolean hasExtendedSpace() { return hasExtendedSpace; }
  public int getExtendedSizeInBytes() { return extendedSizeInBytes; }
  public int getExtendedOffset() { return extendedOffset; }
  public void setExtendedOffset(int value)
  {
    if (value > sizeInBytes) extendedOffset = value;
  }
  @Override
  public Object clone()
  {
    Object o = null;
    try
    {
      o = super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      MsgLog.log("TTaggedStructure", "cannot clone " + e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return o;
  }
  /**
   * Creates a tagged structure with the given name.
   * @param name
   *          is the name to be used as the structure tag. Note that only the
   *          first 16 characters of the given name will be used as the tag
   *          name!
   *          
   * @note The addField() method must be called for each element of the
   *       structure data element.
   *       
   * @note Properties using tagged structures should register themselves
   *       correctly, either by passing the structure name when creating the
   *       TExportProperty() or if using a configuration file such as 
   *       exports.csv or fec.xml, then by supplying the structure name 
   *       there (as in FORMAT -> struct.MYTAG, where "MYTAG" is the 
   *       tagged structure name).
   */
  public TTaggedStructure(String name)
  {
    structName = new String(name);
    fieldArrayList = new ArrayList<structField>();
    if (TStructRegistry.contains(structName))
    {
      registered = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
      TStructRegistry.addToSystem(structName);
    }
    else
      sizeInBytes = 0;
  }
  public void addFieldsFrom(TTaggedStructure tts)
  {
    if (tts != null) 
    {
      Object[] elementData = tts.fieldArrayList.toArray();
      String fieldName;
      for (Object e : elementData) 
      {
        fieldName = ((structField)e).getName();
        if (e instanceof byteArrayField) 
        {
          byte[] field  = new byte[((byteArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof charArrayField)
        {
          char[] field  = new char[((charArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof doubleArrayField)
        {
          double[] field  = new double[((doubleArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof floatArrayField) 
        {
          float[] field  = new float[((floatArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof imageArrayField)
        {
          IMAGE[] field  = new IMAGE[((imageArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof intArrayField)
        {
          int[] field  = new int[((intArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof longArrayField)
        {
          long[] field  = new long[((longArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof shortArrayField)
        {
          short[] field  = new short[((shortArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof spectrumArrayField)
        {
          ASPECTRUM[] field  = new ASPECTRUM[((spectrumArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof StringArrayField) 
        {
          String[] field  = new String[((StringArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof StringBufferField)
        {
          StringBuffer field  = new StringBuffer();
          addField(field,fieldName);
        }
        else if (e instanceof TaggedStructureArrayField)
        {
          try
          {
            TTaggedStructure[] field  = new TTaggedStructure[((TaggedStructureArrayField)e).data.length];
            for (int i=0; i< field.length; i++)
              field[i] = ((TTaggedStructure[])((TaggedStructureArrayField)e).data)[i].getClass().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TaggedStructureField)
        {
          try
          {
            TTaggedStructure field  = ((TaggedStructureField)e).data.getClass().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TCompoundDataArrayField)
        {
          try
          {
            TCompoundDataObject[] field  = new TCompoundDataObject[((TCompoundDataArrayField)e).data.length];
            for (int i=0; i< field.length; i++)
              field[i] = ((TCompoundDataObject[])((TCompoundDataArrayField)e).data)[i].getClass().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TCompoundDataField)
        {
          TCompoundDataObject field;
          try
          {
            field = ((TCompoundDataField)e).data.getClass().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TStringField)
        {
          TString field  = new TString();
          addField(field,fieldName);
        }
      }
      initDone();
    }    
  }
  /**
   * Creates a tagged structure using the class name as the structure
   *        tag. Note that only the first 16 characters of the class name will
   *        be used as the tag name!
   * @note The addField() method must be called for each element of the
   *       structure data element.
   *        
   * \b Example:
   * @include eg_tineTaggedStructure1.java
   * 
   * @note Properties using tagged structures should register themselves
   *       correctly, either by passing the structure name when creating the
   *       TExportProperty() or if using a configuration file such as 
   *       exports.csv or fec.xml, then by supplying the structure name 
   *       there (as in FORMAT -> struct.MYTAG, where "MYTAG" is the 
   *       tagged structure name).
   */
  public TTaggedStructure()
  {
    structName = new String(this.getClass().getSimpleName());
    fieldArrayList = new ArrayList<structField>();
    if (TStructRegistry.contains(structName))
    {
      registered = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
      TStructRegistry.addToSystem(structName);
    }
    else
      sizeInBytes = 0;
  }
  /**
   * Returns the structure's assigned tag name.
   * @return The structure's assigned tag name as a string.
   */
  public String getName()
  {
    return structName;
  }
  /**
   * Returns the structure's data size in bytes.
   * @return The structure's data size in bytes.
   */
  public int getSizeInBytes()
  {
    return sizeInBytes;
  }
  private void resetExtendedSizeInBytes()
  {
    extendedSizeInBytes = 0;
  }
  /**
   * scan thru the extended space and fix the reference and the total size
   */
  public void fixExtendedBytes()
  {
    resetExtendedSizeInBytes();
    for (int i = 0; i < fieldArrayList.size(); i++)
    {
      fieldArrayList.get(i).fixExtendedBytes();
    }    
  }
  public byte[] toByteArray()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(getSizeInBytes());
    baos.reset();
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      structField sf;
      resetExtendedSizeInBytes();
      for (int i = 0; i < fieldArrayList.size(); i++)
      {
        sf = fieldArrayList.get(i);
        sf.fixExtendedBytes();
        sf.writeToOutputStream(ds);
      }
      ds.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure", "---Internal error: cannot convert TaggedStructure: " + getName()
          + " to byte array ", TErrorList.code_failure, e, 0);
      e.printStackTrace();
      return null;
    }
    return baos.toByteArray();
  }
  public byte[] toExtendedByteArray()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(getExtendedSizeInBytes());
    baos.reset();
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      structField sf;
      for (int i = 0; i < fieldArrayList.size(); i++)
      {
        sf = fieldArrayList.get(i);
        sf.writeExtendedBytes(ds);
      }
      ds.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure", "---Internal error: cannot write extended bytes: " + getName()
          + " to byte array ", TErrorList.code_failure, e, 0);
      e.printStackTrace();
      return null;
    }
    return baos.toByteArray();
  }
  public byte[] getByteArray()
  {
    return toByteArray();
  }
  public void toStruct(byte[] data)
  {
    toStruct(data,0,data.length);
  }
  public void toStruct(byte[] data,int offset,int length)
  {
    ByteArrayInputStream dis = new ByteArrayInputStream(data,offset,length);
    DataInputStream ds = new DataInputStream(dis);
    try
    {
      structField sf;
      for (int i = 0; i < fieldArrayList.size(); i++)
      {
        sf = fieldArrayList.get(i);
        sf.readFromInputStream(ds);
        sf.readExtendedBytes(data);
      }
      ds.close();
      dis.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure",
          "---Internal error: cannot convert input data stream to TaggedStructure: " + getName(),
          TErrorList.code_failure, e, 0);
      e.printStackTrace();
    }
  }
  /*
   * Returns a string representation of this tagged structure
   */
  @Override
  synchronized public String toString()
  {
    StringBuffer sb = new StringBuffer("");
    structField sf;
    String indentString = new String("");
    indent += 3;
    for (int i = 0; i < indent; i++)
      indentString += " ";
    sb.append("\n" + indentString + "Struct: " + getName() + "\n");
    sb.append(indentString + "{\n");
    for (int i = 0; i < fieldArrayList.size(); i++)
    {
      sf = (structField) fieldArrayList.get(i);
      sb.append(indentString + " " + sf.getName() + " ->");
      sb.append(sf.toString());
      sb.append("\n");
    }
    sb.append(indentString + "}");
    indent -= 3;
    return sb.toString();
  }
  /**
   * Terminates the tagged structure definition and signals that the
   *        tagged structure is ready for use. This method plays an integral
   *        role in defining the tagged structure. It should be called following
   *        all calls to addField().
   */
  public void initDone()
  {
    if (!registered)
    {
      TStructRegistry.fill(structName, TFormat.CF_NULL, 1);
      registered = true;
    }
    extendedOffset = sizeInBytes;
    TStructDescription tsd = TStructRegistry.get(structName);
    if (tsd != null) tsd.setTaggedStructure(this);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(byte[] field, String fieldName)
  {
    fieldArrayList.add(new byteArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_BYTE, field.length);
      sizeInBytes += field.length;
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(char[] field, String fieldName)
  {
    fieldArrayList.add(new charArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.length);
      sizeInBytes += field.length;
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(short[] field, String fieldName)
  {
    fieldArrayList.add(new shortArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_SHORT, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_SHORT);
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(int[] field, String fieldName)
  {
    fieldArrayList.add(new intArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_LONG, field.length);
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_LONG);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(float[] field, String fieldName)
  {
    fieldArrayList.add(new floatArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_FLOAT, field.length);
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_FLOAT);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(double[] field, String fieldName)
  {
    fieldArrayList.add(new doubleArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_DOUBLE, field.length);
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_DOUBLE);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(long[] field, String fieldName)
  {
    fieldArrayList.add(new longArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_DLONG, field.length);
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_DLONG);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TString field, String fieldName)
  {
    fieldArrayList.add(new TStringField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.getCapacity());
      sizeInBytes += field.getCapacity();
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(StringBuffer field, String fieldName)
  {
    fieldArrayList.add(new StringBufferField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.capacity());
      sizeInBytes += field.capacity();
    }
  }
  public void addField(String[] field, String fieldName)
  {
    fieldArrayList.add(new StringArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_STRING, field.length);
    sizeInBytes += 8 * field.length;
    // has extended bytes
  }
  public void addField(IMAGE[] field, String fieldName)
  {
    fieldArrayList.add(new imageArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_AIMAGE, field.length);
    sizeInBytes += (IMAGE.HEADER_SIZE+8) * field.length;
    // has extended bytes
  }
  public void addField(ASPECTRUM[] field, String fieldName)
  {
    fieldArrayList.add(new spectrumArrayField(field, fieldName));
    if (registered) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_ASPECTRUM, field.length);
    sizeInBytes += (ASPECTRUM.hdrSizeInBytes+8) * field.length;
    // has extended bytes
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TCompoundDataObject field, String fieldName)
  {
    fieldArrayList.add(new TCompoundDataField(field, fieldName));
    if (!registered)
    {
      int num = 1;
      if (TFormat.hasHeader(field.getFormat()))
      {
        int hsiz = TFormat.getFormatHeaderSize(field.getFormat());
        num = (field.getSizeInBytes() - hsiz) / TFormat.formatSizeOf(field.getFormat());
      }
      TStructRegistry.fill(structName, fieldName, field.getFormat(), num);
      sizeInBytes += field.getSizeInBytes();
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TCompoundDataObject[] field, String fieldName)
  {
    fieldArrayList.add(new TCompoundDataArrayField(field, fieldName));
    if (!registered)
    {
      Object tdo = field[0];
      if (field[0] == null) try
      {
        tdo = field.getClass().getComponentType().newInstance();
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("addField", "cannot add field : "+e.getMessage(),TErrorList.code_failure,e,0);
        return;
      }
      TStructRegistry.fill(structName, fieldName, ((TCompoundDataObject) tdo).getFormat(), field.length);
      sizeInBytes += ((TCompoundDataObject) tdo).getSizeInBytes() * field.length;
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TTaggedStructure field, String fieldName)
  {
    int size = TStructRegistry.getSizeInBytes(field.getName());
    if (size < 1)
    {
      MsgLog.log("TTaggedStructure", "---Error: attempt to register a field: " + fieldName
          + " of unknown data type: " + field.getName(), TErrorList.code_failure, null, 0);
      return;
    }
    fieldArrayList.add(new TaggedStructureField(field, fieldName));
    if (!registered)
    {
      String fieldNameTag = "<" + field.getName() + ">" + fieldName;
      TStructRegistry.fill(structName, fieldNameTag, TFormat.CF_STRUCT, 1);
      sizeInBytes += size;
    }
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TTaggedStructure[] field, String fieldName)
  {
    int size = TStructRegistry.getSizeInBytes(field[0].getName());
    if (size < 1)
    {
      MsgLog.log("TTaggedStructure", "attempt to register a field: " + fieldName
          + " being an array of unknown data type: " + field[0].getName(), TErrorList.code_failure, null, 0);
      return;
    }
    //size = size * field.length; << why was this ever here ?
    fieldArrayList.add(new TaggedStructureArrayField(field, fieldName));
    if (!registered)
    {
      String fieldNameTag = "<" + field[0].getName() + ">" + fieldName;
      TStructRegistry.fill(structName, fieldNameTag, TFormat.CF_STRUCT, field.length);
      sizeInBytes += size * field.length;
    }
  }
  /*
   * General Byte Alignment Rules -------Taken from:
   * http://www.eventhelix.com/RealtimeMantra/ByteAlignmentAndOrdering.htm The
   * following byte padding rules will generally work with most 32 bit
   * processor. You should consult your compiler and microprocessor manuals to
   * see if you can relax any of these rules. - Single byte numbers can be
   * aligned at any address - Two byte numbers should be aligned to a two byte
   * boundary - Four byte numbers should be aligned to a four byte boundary -
   * Structures between 1 and 4 bytes of data should be padded so that the total
   * structure is 4 bytes. - Structures between 5 and 8 bytes of data should be
   * padded so that the total structure is 8 bytes. - Structures between 9 and
   * 16 bytes of data should be padded so that the total structure is 16 bytes.
   * - Structures greater than 16 bytes should be padded to 16 byte boundary.
   * Karol's comment: it seems not to work for Linux GCC... For now I decide to
   * limit the padding to 4 bytes
   */
  // private void checkPadding()
  // {
  // return;
  // // if( offset < 1 ) return;
  // //
  // // int n = offset % 4;
  // // if( n == 0 ) return;
  // // addField( new byte[4-n], "reserved");
  // }
  /*
   * private void checkPadding_not_working() { if( offset < 1 ) return; if(
   * offset < 4 ) { addField( new byte[ 4 - offset ], "reserved"); return; } if(
   * offset < 8 ) { addField( new byte[ 8 - offset ], "reserved"); return; } if(
   * offset < 16 ) { addField( new byte[ 16 - offset ], "reserved"); return; }
   * int n = offset % 16; if( n == 0 ) return; addField( new byte[16-n],
   * "reserved"); }
   */
  /*
   * old version: private int padding = 0; private void checkAligment(int size)
   * { //check the size for the correct alignment. Add some placeholders if
   * needed int d = size % 4; int fieldPadding = 4 - d; if( d > 0 ) { // there
   * are extra padding bytes if (padding > 0 && d > padding) { // too many bytes
   * to append to current padding set so clear it padding = fieldPadding;
   * addField( new byte[padding], "reserved"); } if (padding == 0) padding = 4;
   * padding -= d; } else { // new field is aligned; are there padding bytes? if
   * (padding > 0 && padding < 4) { padding = fieldPadding; addField( new
   * byte[padding], "reserved"); } } }
   */
  // ----------------------------------------------
  private abstract class structField
  {
    private String name;
    public String getName()
    {
      return name;
    }
    public void setName(String name)
    {
      this.name = name;
    }
    abstract void writeToOutputStream(DataOutputStream ds) throws IOException;
    abstract void readFromInputStream(DataInputStream ds) throws IOException;
    abstract void fixExtendedBytes();
    abstract void readExtendedBytes(byte[] bytes) throws IOException;
    abstract void writeExtendedBytes(DataOutputStream ds) throws IOException;
    @Override
    abstract public String toString();
  }
  private class byteArrayField extends structField
  {
    byte[] data;
    byteArrayField(byte[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      ds.read(data, 0, data.length);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data, 0, data.length);
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      int ibyte;
      for (int i = 0; i < data.length; i++)
      {
        if ((ibyte = data[i]) < 0) ibyte += 256;
        sb.append(" " + ibyte);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class charArrayField extends structField
  {
    char[] data;
    charArrayField(char[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = (char) (ds.readByte());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeByte((byte) (data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length && data[i] != 0; i++)
      {
        sb.append(data[i]);
      }
      return sb.toString().trim();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class shortArrayField extends structField
  {
    short[] data;
    shortArrayField(short[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = Swap.Short(ds.readShort());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeShort(Swap.Short(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class intArrayField extends structField
  {
    int[] data;
    intArrayField(int[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = Swap.Long(ds.readInt());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeInt(Swap.Long(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class floatArrayField extends structField
  {
    float[] data;
    floatArrayField(float[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[4];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.Float(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.write(Swap.Float(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class doubleArrayField extends structField
  {
    double[] data;
    doubleArrayField(double[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[8];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.Double(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.write(Swap.Double(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class longArrayField extends structField
  {
    long[] data;
    longArrayField(long[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[8];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.DLong(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeLong(Swap.DLong(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TStringField extends structField
  {
    TString data;
    TStringField(TString data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] b = new byte[data.getCapacity()];
      ds.read(b);
      data.set(b);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.getBytes());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class StringBufferField extends structField
  {
    StringBuffer data;
    StringBufferField(StringBuffer data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.capacity(); i++)
        data.setCharAt(i, (char) ds.readByte());
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.capacity(); i++)
        ds.writeByte((byte) (data.charAt(i)));
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class StringArrayField extends structField
  {
    String[] data;
    int[] xoff;
    int[] xsiz;
    StringArrayField(String[] data, String name)
    {
      setName(name);
      this.data = data;
      if (data == null) return;
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        xoff[i] = Swap.Long(ds.readInt());
        xsiz[i] = Swap.Long(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        data[i] = new String(bytes,xoff[i],xsiz[i]-1);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.writeInt(Swap.Long(xoff[i]));
        ds.writeInt(Swap.Long(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? 0 : data[i].length() + 1;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        if (data[i] != null) ds.write(data[i].getBytes());
        ds.write(0);
      }
    }
    @Override
    public String toString()
    {
      String indentString = "\n    "; 
      for (int i =0; i < getName().length(); i++)
        indentString += " ";
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        if (i > 0) sb.append(indentString + " ->");        
        sb.append(" "+data[i]);
      }
      return sb.toString();
    }
  }
  private class TCompoundDataField extends structField
  {
    TCompoundDataObject data;
    TCompoundDataField(TCompoundDataObject data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data instanceof IMAGE)
      {
        ((IMAGE) data).read(ds);
      }
      else
      {
        ds.read(data.getByteArray());
        data.toStruct();
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.toByteArray());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TCompoundDataArrayField extends structField
  {
    TCompoundDataObject[] data;
    TCompoundDataArrayField(TCompoundDataObject[] data, String name)
    {
      setName(name);
      this.data = data;
      for (int i=0; i<data.length; i++)
      { // make sure something is instantiated
        if (((TCompoundDataObject[]) data)[i] == null)
        { // wasn't there -> make one
          try
          {
            data[i] = (TCompoundDataObject) data.getClass().getComponentType().newInstance();
          }
          catch (Exception e)
          {
            e.printStackTrace();
            MsgLog.log("makeNewCompoundDataObject", e.toString(), TErrorList.code_failure, e, 0);
          }        
        }
      }
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (TCompoundDataObject obj : data)
      {
        if (obj instanceof IMAGE)
        {
          ((IMAGE) obj).read(ds);
        }
        else
        {
          ds.read(obj.getByteArray());
          obj.toStruct();
        }
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
        ds.write(data[i].toByteArray());
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
        sb.append(" " + data[i]);
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TaggedStructureField extends structField
  {
    TTaggedStructure data;
    TaggedStructureField(TTaggedStructure data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[data.getSizeInBytes()];
      ds.read(buffer);
      data.toStruct(buffer);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.toByteArray());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TaggedStructureArrayField extends structField
  {
    TTaggedStructure[] data;
    TaggedStructureArrayField(TTaggedStructure[] data, String name)
    {
      setName(name);
      this.data = data;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[data[0].getSizeInBytes()];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i].toStruct(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
        ds.write(data[i].toByteArray());
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
        sb.append(" " + data[i]);
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class imageArrayField extends structField
  {
    int[] xoff;
    int[] xsiz;
    IMAGE[] data;
    imageArrayField(IMAGE[] data, String name)
    {
      setName(name);
      this.data = data;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        if (data[i] == null) data[i] = new IMAGE(0x10000);
      }
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
    }
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        data[i].readHeaders(ds);
        xoff[i] = Swap.Long(ds.readInt());
        xsiz[i] = Swap.Long(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        if (xsiz[i] + xoff[i] > bytes.length) 
        {
          MsgLog.log("TTaggedStructure.readExtendedBytes", "extended byte information is corrupt",TErrorList.code_failure,null,1);
          return;
        }
        data[i].ensureFrameBufferCapacity(xsiz[i]);
        data[i].fillFrameBuffer(bytes, xoff[i], 0, xsiz[i]);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        data[i].writeHeaders(ds);
        ds.writeInt(Swap.Long(xoff[i]));
        ds.writeInt(Swap.Long(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? 0 : data[i].getFrameHeader().appendedFrameSize;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].getImageFrameBuffer(),0,xsiz[i]);
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i].toString());
      }
      return sb.toString();
    }
    
  }
  private class spectrumArrayField extends structField
  {
    int[] xoff;
    int[] xsiz;
    ASPECTRUM[] data;
    spectrumArrayField(ASPECTRUM[] data, String name)
    {
      setName(name);
      this.data = data;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        if (data[i] == null) data[i] = new ASPECTRUM(TFormat.SPECTRUM_STD_SIZE);
      }
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
    }
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.read(data[i].getByteArray());
        data[i].toStruct();
        xoff[i] = Swap.Long(ds.readInt());
        xsiz[i] = Swap.Long(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      byte[] ba;
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        if (xsiz[i] + xoff[i] > bytes.length) 
        {
          MsgLog.log("TTaggedStructure.readExtendedBytes", "extended byte information is corrupt",TErrorList.code_failure,null,1);
          return;
        }
        ba = new byte[xsiz[i]];
        System.arraycopy(bytes, xoff[i], ba, 0, xsiz[i]);
        data[i].setSpectrumByteArray(ba);
        data[i].toSpectrumStruct();
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].toByteArray());
        ds.writeInt(Swap.Long(xoff[i]));
        ds.writeInt(Swap.Long(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? TFormat.SPECTRUM_STD_SIZE : data[i].getLength() * 4;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].toSpectrumByteArray(),0,xsiz[i]);
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        sb.append(" " + data[i].toString());
      }
      return sb.toString();
    }
    
  }
}
